from typing import Optional, Any

from sai_rl.model.handlers._base import BaseModelHandler
from sai_rl.sai_console import SAIConsole

import gymnasium as gym
import numpy as np

IMPORT_ERROR_MESSAGE = (
    "TensorFlow is not installed. Please install it using 'pip install sai_rl[tf]'."
)


class KerasModelHandler(BaseModelHandler):
    _tf: Any = None
    _keras: Any = None
    _tf2onnx: Any = None
    _onnx: Any = None
    _tf_available: bool = False

    def __init__(self, env: gym.Env, console: Optional[SAIConsole] = None):
        super().__init__(env, console)
        try:
            import tensorflow as tf
            import keras
            import tf2onnx
            import onnx

            self._tf = tf
            self._keras = keras
            self._tf2onnx = tf2onnx
            self._onnx = onnx
            self._tf_available = True
            pass
        except ImportError:
            self._tf_available = False
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def _check_tensorflow_available(self) -> None:
        if not self._tf_available:
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def load_model(self, model_path):
        self._check_tensorflow_available()
        return self._keras.models.load_model(model_path)

    def get_policy(self, model, obs: np.ndarray) -> np.ndarray:
        self._check_tensorflow_available()

        if len(obs.shape) < self.obs_space_dims + 1:
            obs = np.expand_dims(obs, axis=0)

        state = self._tf.convert_to_tensor(obs, dtype=self.observation_space.dtype)
        policy = model(state)
        return np.array(policy)

    def save_model(self, model, model_path):
        self._check_tensorflow_available()
        inference_model = self.convert_to_sequential(model)
        inference_model.save(model_path)

    def convert_to_sequential(self, model):
        inference_model = self._keras.Sequential(
            [
                self._keras.layers.InputLayer(shape=(model.n_features,)),
                *model.hidden_layers,
                model.output_layer,
            ]
        )

        for src_layer, dst_layer in zip(model.layers, inference_model.layers):
            dst_layer.set_weights(src_layer.get_weights())

        return inference_model

    def hash(self, model):
        formatted_model = self.convert_to_sequential(model)
        return {
            "architecture_hash": self._hash(formatted_model.to_json()),
            "parameters_hash": self._hash(
                [w.tolist() for w in formatted_model.get_weights()]
            ),
        }

    def export_to_onnx(self, model, model_path: str):
        self._check_tensorflow_available()

        model.output_names = ["output"]
        input_signature = [
            self._tf.TensorSpec(
                [None, *self.observation_space.sample().shape],
                self._tf.float32,
                name="input",
            )
        ]
        onnx_model, _ = self._tf2onnx.convert.from_keras(
            model, input_signature=input_signature, opset=18
        )
        self._onnx.save(
            onnx_model,
            model_path if model_path.endswith(".onnx") else f"{model_path}.onnx",
        )



from typing import Optional, TypedDict, Literal
import numpy as np


class MatchResults(TypedDict):
    score: float
    duration: float
    videoKey: Optional[str]


class BenchmarkResults(TypedDict):
    status: Literal["success", "error", "timeout"]
    score: Optional[np.float64]
    duration: float
    matches: list[MatchResults]
    logs: Optional[str]
    error: Optional[str]



from typing import Optional, TypedDict, List
from types import SimpleNamespace
import json

from sai_rl.api.routes.package import PackageType

from sai_rl.api.requestor import APIRequestor

from sai_rl.types import EnvironmentStandardType, EnvironmentActionType


class EnvironmentType(TypedDict):
    id: str
    name: str
    gymId: str
    type: EnvironmentStandardType
    actionType: EnvironmentActionType
    package: PackageType


class EnvironmentAPI:
    def __init__(self, api: APIRequestor):
        self._api = api

    def get(self, environment_id: str) -> Optional[EnvironmentType]:
        response = self._api.get(f"/v1/environments/{environment_id}")

        if not response:
            return None

        raw_environment = response.json()
        if not raw_environment:
            return None

        env_obj = json.loads(json.dumps(raw_environment), object_hook=lambda d: SimpleNamespace(**d))
        environment: EnvironmentType = {
            "id": env_obj.id,
            "name": env_obj.name,
            "gymId": env_obj.gymId,
            "type": env_obj.type,
            "actionType": env_obj.actionType,
            "package": {
                "id": env_obj.package.id,
                "name": env_obj.package.name,
                "version": env_obj.package.version
            }
        }

        return environment

    def list(self) -> Optional[List[EnvironmentType]]:
        response = self._api.get("/v1/environments")

        if not response:
            return None

        raw_environments = response.json()
        if not raw_environments:
            return None

        environments: List[EnvironmentType] = []

        for raw_environment in raw_environments:
            env_obj = json.loads(json.dumps(raw_environment), object_hook=lambda d: SimpleNamespace(**d))
            environment: EnvironmentType = {
                "id": env_obj.id,
                "name": env_obj.name,
                "gymId": env_obj.gymnasiumEnv,
                "type": env_obj.type,
                "actionType": env_obj.actionType,
                "package": {
                    "id": env_obj.package.id,
                    "name": env_obj.package.name,
                    "version": env_obj.package.version
                }
            }

            environments.append(environment)

        return environments



from typing import Optional, TypedDict, List
from types import SimpleNamespace
import json

from sai_rl.api.requestor import APIRequestor
from sai_rl.types import EnvironmentActionType, EnvironmentStandardType


class SubmissionEnvironmentType(TypedDict):
    id: str
    name: str
    gymId: str
    type: EnvironmentStandardType
    actionType: EnvironmentActionType


class SubmissionCompetitionType(TypedDict):
    id: str
    slug: str
    name: str
    environment: SubmissionEnvironmentType


class SubmissionType(TypedDict):
    id: str
    name: str
    type: str
    status: str
    score: Optional[str]
    competition: SubmissionCompetitionType
    updatedAt: str
    createdAt: str


class SubmissionAPI:
    def __init__(self, api: APIRequestor):
        self._api = api

    def list(self) -> Optional[List[SubmissionType]]:
        response = self._api.get("/v1/submissions")

        if not response:
            return None

        raw_submissions = response.json()
        if not raw_submissions:
            return None

        submissions: List[SubmissionType] = []

        for raw_submission in raw_submissions:
            sub_obs = json.loads(
                json.dumps(raw_submission), object_hook=lambda d: SimpleNamespace(**d)
            )
            submission: SubmissionType = {
                "id": sub_obs.id,
                "name": sub_obs.name,
                "type": sub_obs.type,
                "status": sub_obs.status,
                "score": sub_obs.score if hasattr(sub_obs, "score") else None,
                "competition": {
                    "id": sub_obs.competition.id,
                    "slug": sub_obs.competition.slug,
                    "name": sub_obs.competition.name,
                    "environment": {
                        "id": sub_obs.competition.environment.id,
                        "name": sub_obs.competition.environment.name,
                        "gymId": sub_obs.competition.environment.gymId,
                        "type": sub_obs.competition.environment.type,
                        "actionType": sub_obs.competition.environment.actionType,
                    },
                },
                "updatedAt": sub_obs.updatedAt,
                "createdAt": sub_obs.createdAt,
            }

            submissions.append(submission)

        return submissions

    def create(self, data: dict, files: dict) -> Optional[SubmissionType]:
        response = self._api.post("/v1/submissions", data=data, files=files)

        if not response:
            return None

        raw_submission = response.json()

        sub_obs = json.loads(
            json.dumps(raw_submission), object_hook=lambda d: SimpleNamespace(**d)
        )
        submission: SubmissionType = {
            "id": sub_obs.id,
            "name": sub_obs.name,
            "type": sub_obs.type,
            "status": sub_obs.status,
            "score": sub_obs.score if hasattr(sub_obs, "score") else None,
            "competition": {
                "id": sub_obs.competition.id,
                "slug": sub_obs.competition.slug,
                "name": sub_obs.competition.name,
                "environment": {
                    "id": sub_obs.competition.environment.id,
                    "name": sub_obs.competition.environment.name,
                    "gymId": sub_obs.competition.environment.gymId,
                    "type": sub_obs.competition.environment.type,
                    "actionType": sub_obs.competition.environment.actionType,
                },
            },
            "updatedAt": sub_obs.updatedAt,
            "createdAt": sub_obs.createdAt,
        }

        return submission



from abc import ABC, abstractmethod
from typing import Optional

import numpy as np
import gymnasium as gym
import json
import hashlib

from sai_rl.sai_console import SAIConsole


class BaseModelHandler(ABC):
    """Base class for model handlers in SAI RL.
    This class provides an interface for loading, saving, and exporting models,
    as well as getting actions from the model.
    """

    _console: Optional[SAIConsole] = None

    is_continuous: bool
    obs_space_dims: int
    action_space: gym.Space
    observation_space: gym.Space

    def __init__(self, env: gym.Env, console: Optional[SAIConsole] = None):
        self._console = console

        self.obs_space_dims = len(env.observation_space.sample().shape)
        self.is_continuous = isinstance(
            env.action_space, gym.spaces.Box
        ) and not np.issubdtype(env.action_space.dtype, np.integer)
        self.action_space = env.action_space
        self.observation_space = env.observation_space

    @abstractmethod
    def load_model(self, model_path: str):
        pass

    @abstractmethod
    def get_policy(self, model, obs: np.ndarray) -> np.ndarray:
        pass

    def get_action(self, model, obs: np.ndarray) -> np.ndarray:
        """Get action from the model given an observation.
        For discrete action spaces, returns argmax of policy distribution.
        For continuous action spaces, returns the policy values, which are assumed to be already scaled
        to the appropriate action range.
        """
        policy = self.get_policy(model, obs)
        if self.is_continuous:
            if np.all(np.isfinite(self.action_space.low)) and np.all(  # type: ignore
                np.isfinite(self.action_space.high)  # type: ignore
            ):
                expected_bounds = [-1, 1]
                action_percent = (policy - expected_bounds[0]) / (
                    expected_bounds[1] - expected_bounds[0]
                )
                bounded_percent = np.minimum(np.maximum(action_percent, 0), 1)
                return (
                    self.action_space.low  # type: ignore
                    + (self.action_space.high - self.action_space.low) * bounded_percent  # type: ignore
                )
            return policy

        if len(policy.shape) == 2:
            return np.argmax(policy, axis=1)

        return policy

    @abstractmethod
    def save_model(self, model, model_path: str):
        pass

    @abstractmethod
    def export_to_onnx(self, model, path: str):
        pass

    @abstractmethod
    def hash(self, model):
        return {}

    def _hash(self, model_json: str):
        serialized_str = json.dumps(model_json, sort_keys=True)
        return hashlib.sha256(serialized_str.encode("utf-8")).hexdigest()



from typing import Optional

import os

from sai_rl.sai_console import SAIConsole
from sai_rl.error import AuthenticationError, ConfigurationError
from sai_rl.utils import config


class RequestorOptions(object):
    def __init__(
        self,
        console: SAIConsole,
        api_key: Optional[str] = None,
        api_base: Optional[str] = None,
        max_network_retries: Optional[int] = None,
    ):
        self._console = console
        self._api_key = api_key
        self._api_base = self._get_api_base(api_base)
        self._validate_api_base(self._api_base or "")

        self._max_network_retries = self._get_max_network_retries(max_network_retries)

    @property
    def api_key(self) -> str:
        if self._api_key is not None:
            return self._api_key
        elif os.getenv("SAI_API_KEY") is not None:
            self._api_key = os.getenv("SAI_API_KEY")
        else:
            self._api_key = config.api_token

        if not self._api_key:
            raise AuthenticationError("""
No API key provided.

Please run 'sai login' to authenticate with your user.
                """)

        return self._api_key

    @staticmethod
    def _get_api_base(api_base: Optional[str] = None) -> Optional[str]:
        if api_base:
            return api_base

        env_api_base = os.getenv("SAI_API_BASE")
        if env_api_base:
            return env_api_base

        return config.api_url

    @staticmethod
    def _validate_api_base(api_base: str):
        if not api_base.startswith("http"):
            raise ConfigurationError("API base must start with 'http'")

    @property
    def api_base(self) -> str:
        assert self._api_base is not None
        return self._api_base

    @staticmethod
    def _get_max_network_retries(
        max_network_retries: Optional[int] = None,
    ) -> Optional[int]:
        if max_network_retries:
            return max_network_retries

        return config.max_network_retries

    @property
    def max_network_retries(self) -> int:
        assert self._max_network_retries is not None
        return self._max_network_retries

    def to_dict(self):
        """
        Returns a dict representation of the object.
        """
        return {
            "api_key": self.api_key,
            "api_base": self.api_base,
            "max_network_retries": self.max_network_retries,
        }



import time
from typing import Optional, Callable

import os
import re
import inspect
import requests

import numpy as np
import gymnasium as gym

from rich.syntax import Syntax

from sai_rl.utils import config
from sai_rl.sai_console import SAIConsole, SAIStatus
from sai_rl.error import ActionFunctionError, NetworkError


class ActionFunctionManager:
    _console: Optional[SAIConsole]
    _env: Optional[gym.Env]

    def __init__(
        self,
        action_function: Optional[str | Callable] = None,
        download_dir: str = config.temp_path,
        env: Optional[gym.Env] = None,
        verbose: bool = False,
        console: Optional[SAIConsole] = None,
        status: Optional[SAIStatus] = None,
    ):
        self._console = console
        self._env = env

        self.verbose = verbose

        self._action_function = None
        self._action_function_source = None

        self._download_dir = download_dir
        self._path = None

        if status:
            status.update("Loading action function...")

        self.load(action_function, status)

    def _print(self):
        if not self._action_function_source:
            return

        if not self._console:
            return

        panel_group = self._console.group(
            Syntax(self._action_function_source, "python", theme="github-dark"),
        )

        panel = self._console.panel(
            panel_group, title="Action Function", padding=(1, 2)
        )

        self._console.print()
        self._console.print(panel)

    def _load_from_code(self, code: str, status: Optional[SAIStatus] = None):
        if status:
            status.update("Loading action function from code...")

        clean_code = self.remove_imports(code)

        function_name = None
        for line in clean_code.splitlines():
            if line.strip().startswith("def "):
                function_name = line.strip().split("def ")[1].split("(")[0].strip()
                break

        if not function_name:
            raise ActionFunctionError(
                "No function definition found in the provided code"
            )

        namespace = dict({"np": np, "env": self._env})
        exec(clean_code, namespace)

        self._action_function = namespace[function_name]
        self._action_function_source = clean_code

    def _load_from_file(self, path: str, status: Optional[SAIStatus] = None):
        if status:
            status.update("Loading action function from file...")

        self._path = path
        if not os.path.exists(path):
            raise ActionFunctionError(f"File not found: {path}")

        with open(path, "r") as f:
            code = f.read()
            self._load_from_code(code)

    def _load_from_url(self, url: str, status: Optional[SAIStatus] = None):
        if status:
            status.update("Downloading action function from URL...")
            status.stop()

        if not self._download_dir:
            raise ActionFunctionError("Download path not set")

        os.makedirs(self._download_dir, exist_ok=True)
        action_path = f"{self._download_dir}/{time.time()}.py"

        if os.path.exists(action_path):
            os.remove(action_path)

        try:
            if self._console:
                with self._console.progress("Downloading action function") as progress:
                    with requests.get(url, stream=True) as response:
                        response.raise_for_status()
                        total_size = int(response.headers.get("content-length", 0))
                        task = progress.add_task("Downloading...", total=total_size)

                        chunk_size = 8192  # 8 KB
                        downloaded_size = 0

                        with open(action_path, "wb") as f:
                            for chunk in response.iter_content(chunk_size=chunk_size):
                                if chunk:
                                    f.write(chunk)
                                    downloaded_size += len(chunk)
                                    progress.update(task, advance=len(chunk))
            else:
                with requests.get(url) as response:
                    response.raise_for_status()
                    with open(action_path, "wb") as f:
                        f.write(response.content)

        except requests.exceptions.RequestException as e:
            raise NetworkError(f"Failed to download action function: {e}")

        if status:
            status.start()

        self._load_from_file(action_path, status)

    @staticmethod
    def remove_imports(code_string: str) -> str:
        """Remove import statements from code for security."""
        pattern = re.compile(
            r"^\s*#?\s*(from\s+\w+\s+import\s+.*|import\s+\w+.*|from\s+\w+\s+import\s+\(.*\)|import\s+\(.*\))",
            re.MULTILINE,
        )
        return re.sub(pattern, "", code_string)

    def get_action(self, obs: np.ndarray) -> np.ndarray:
        if self._action_function is None:
            raise ActionFunctionError("Action function not loaded")

        return self._action_function(obs)  # type: ignore

    def load(
        self,
        action_fn: Optional[str | Callable] = None,
        status: Optional[SAIStatus] = None,
    ):
        if status:
            status.update("Loading action function...")

        if not action_fn:
            if self._console:
                self._console.warning("No action function provided, skipping load.")
            return

        if isinstance(action_fn, str):
            if action_fn.startswith(("http://", "https://")):
                self._load_from_url(action_fn, status)
            elif action_fn.endswith((".py")) and os.path.exists(action_fn):
                self._load_from_file(action_fn, status)
            elif action_fn.startswith("def"):
                self._load_from_code(action_fn, status)
            else:
                raise ActionFunctionError(
                    f"Unsupported action function path: {action_fn}"
                )
        elif callable(action_fn):
            action_fn_source = inspect.getsource(action_fn)
            self._load_from_code(action_fn_source)
        else:
            raise ActionFunctionError(
                f"Unsupported action function type: {type(action_fn)}"
            )

        if self.verbose:
            self._print()

        if self._console:
            self._console.success("Successfully loaded action function.")

    def save_action_function(self, path: str):
        if not self._action_function_source:
            raise ActionFunctionError("No action function loaded")

        if not os.path.exists(os.path.dirname(path)):
            os.makedirs(os.path.dirname(path), exist_ok=True)

        if os.path.exists(path):
            raise ActionFunctionError(f"File already exists: {path}")

        if not path.endswith(".py"):
            raise ActionFunctionError("File must end with .py")

        with open(path, "w") as f:
            f.write(self._action_function_source)
        self._path = path

    def clean(self):
        if self._path and os.path.exists(self._path):
            os.remove(self._path)
            self._path = None
        self._action_function = None
        self._action_function_source = None



from typing import Any, TYPE_CHECKING, Optional
from sai_rl.model.handlers._base import BaseModelHandler

from sai_rl.sai_console import SAIConsole

import json
import numpy as np
import gymnasium as gym

if TYPE_CHECKING:
    import torch
    import torch.nn as nn

    PyTorchModel = nn.Module
    PyTorchTensor = torch.Tensor
else:
    PyTorchModel = Any
    PyTorchTensor = Any

IMPORT_ERROR_MESSAGE = (
    "PyTorch is not installed. Please install it using 'pip install sai_rl[torch]'."
)


class PyTorchModelHandler(BaseModelHandler):
    _torch: Any = None
    _torch_available: bool = False

    _ignored_keys = {
        '_non_persistent_buffers_set',
        '_parameters',
        '_buffers',
        '_modules',
        '_backward_pre_hooks',
        '_backward_hooks',
        '_is_full_backward_hook',
        '_forward_hooks',
        '_forward_hooks_with_kwargs',
        '_forward_hooks_always_called',
        '_forward_pre_hooks',
        '_forward_pre_hooks_with_kwargs',
        '_state_dict_hooks',
        '_state_dict_pre_hooks',
        '_load_state_dict_pre_hooks',
        '_load_state_dict_post_hooks',
    }    

    def __init__(self, env: gym.Env, console: Optional[SAIConsole] = None):
        super().__init__(env, console)
        try:
            import torch

            self._torch = torch
            self._torch_available = True
        except ImportError:
            self._torch_available = False
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def _check_torch_available(self) -> None:
        if not self._torch_available:
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def _safe_config(self, module):
        config = {}

        for k, v in vars(module).items():
            if k in {"_parameters", "_buffers", "_modules"}:
                continue
            if k in self._ignored_keys:
                continue

            if isinstance(v, (self._torch.nn.Parameter, self._torch.Tensor, self._torch.nn.Module)):
                continue
            elif isinstance(v, float):
                config[k] = round(v, 8)

            try:
                if callable(v):
                    config[k] = v.__name__
                else:
                    json.dumps(v)
                    config[k] = v
            except TypeError:
                config[k] = str(v)

        return config

    def _serialize_architecture(self, model):
        def serialize_module(m: self._torch.nn.Module):
            return {
                "class": m.__class__.__name__,
                "config": self._safe_config(m),
                "children": [serialize_module(child) for child in m.children()],
            }

        return serialize_module(model)

    def load_model(self, model_path: str, use_torchscript: bool = True) -> PyTorchModel:
        """
        Load a PyTorch model from the specified path.
        Args:
            model_path (str): Path to the model file.
            use_torchscript (bool): If True, load a TorchScript model. Otherwise, load a regular PyTorch model.
        Returns:
            torch.nn.Module: The loaded PyTorch model.
        Raises:
            ImportError: If PyTorch is not installed.
        """
        self._check_torch_available()

        if use_torchscript:
            model = self._torch.jit.load(model_path)
        else:
            model = self._torch.load(model_path)

        model.eval()
        return model

    def get_policy(self, model: PyTorchModel, obs: np.ndarray) -> np.ndarray:
        """
        Get the action from the model given an observation.
        Args:
            model (torch.nn.Module): The PyTorch model.
            obs (np.ndarray): The observation input to the model.
            is_continuous (bool): If True, the model is for continuous action space.
        Returns:
            int: The action selected by the model.
        Raises:
            ImportError: If PyTorch is not installed.
        """
        self._check_torch_available()

        if len(obs.shape) < self.obs_space_dims + 1:
            obs = np.expand_dims(obs, axis=0)

        obs_tensor = self._torch.from_numpy(obs)
        return model(obs_tensor).detach().numpy()

    def save_model(
        self, model: PyTorchModel, model_path: str, use_torchscript: bool = True
    ):
        """
        Save a PyTorch model to the specified path.
        Args:
            model (torch.nn.Module): The PyTorch model to save.
            model_path (str): Path to save the model file.
            use_torchscript (bool): If True, save as a TorchScript model. Otherwise, save as a regular PyTorch model.
        Returns:
            str: The path to the saved model file.
        Raises:
            ImportError: If PyTorch is not installed.
        """
        self._check_torch_available()

        if use_torchscript:
            if not isinstance(model, self._torch.jit.ScriptModule):
                obs_tensor = self._torch.from_numpy(
                    np.zeros(
                        (1, *self.observation_space.shape),
                        dtype=self.observation_space.dtype,
                    )
                )

                model = self._torch.jit.trace(
                    model,
                    self._torch.randn(
                        *obs_tensor.shape,
                        dtype=obs_tensor.dtype,
                    ),
                )
            self._torch.jit.save(model, model_path)
        else:
            self._torch.save(model, model_path)
        return model_path

    def hash(self, model):
        return {
            "architecture_hash": self._hash(self._serialize_architecture(model)),
            "parameters_hash": self._hash([param.detach().cpu().tolist() for param in model.parameters()])
        }        

    def export_to_onnx(self, model: PyTorchModel, model_path: str):
        self._check_torch_available()
        model.eval()

        torch_input = self._torch.from_numpy(
            np.zeros(
                (1, *self.observation_space.shape),
                dtype=self.observation_space.dtype,
            )
        )

        return self._torch.onnx.export(
            model,
            (torch_input,),
            model_path,
            export_params=True,
            opset_version=18,
            input_names=["input"],
            output_names=["output"],
            dynamic_axes={"input": {0: "batch_size"}, "output": {0: "batch_size"}},
        )



import gymnasium as gym
from typing import Optional

from sai_rl.error import ModelError
from sai_rl.types import ModelLibraryType

from sai_rl.model.handlers._base import BaseModelHandler

from sai_rl.sai_console import SAIConsole


def get_handler(
    env: gym.Env,
    model_type: ModelLibraryType,
    algorithm: Optional[str],
    console: Optional[SAIConsole],
) -> BaseModelHandler:
    if model_type == "pytorch":
        from sai_rl.model.handlers.pytorch import PyTorchModelHandler

        handler = PyTorchModelHandler(env, console)
    elif model_type == "tensorflow":
        from sai_rl.model.handlers.tensorflow import TensorFlowModelHandler

        handler = TensorFlowModelHandler(env, console)
    elif model_type == "keras":
        from sai_rl.model.handlers.keras import KerasModelHandler

        handler = KerasModelHandler(env, console)
    elif model_type == "stable_baselines3":
        from sai_rl.model.handlers.sb3 import SBL3ModelHandler

        if algorithm is None:
            raise ModelError(
                "Attempted to load a Stable Baselines3 model without specifying an algorithm. "
                "Please specify 'algorithm' parameter explicitly. "
                "Supported algorithms are: A2C, DDPG, DQN, HER, PPO, SAC, TD3"
            )
        handler = SBL3ModelHandler(env, console, algorithm)
    elif model_type == "onnx":
        from sai_rl.model.handlers.onnx import OnnxModelHandler

        handler = OnnxModelHandler(env, console)
    else:
        raise ModelError(f"Unsupported model type: {model_type}")

    return handler


__all__ = ["get_handler", "BaseModelHandler"]



from typing import Optional, Callable

import os
import requests
import gymnasium as gym

from rich.align import Align
from rich.text import Text

from sai_rl.model.action_manager import ActionFunctionManager
from sai_rl.sai_console import SAIConsole, SAIStatus

from sai_rl.utils import config
from sai_rl.error import ModelError, NetworkError
from sai_rl.types import ModelType, ModelLibraryType

from sai_rl.model.handlers import get_handler, BaseModelHandler


"""
For model hashing, we only support TensorFlow, ONNX, and Stable Baselines3 models.

The PyTorch hashing is less than optimal, so we can skip it for now until we have a better solution.

Keras hashing is not implemented, as it is not a supported model type for the platform.
"""
SUPPORTED_HASHING_MODEL_TYPES = ["tensorflow", "onnx", "stable_baselines3"]


def get_model_suffix(model_type: ModelLibraryType) -> str:
    """Get the appropriate file suffix for a model type."""
    suffixes = {
        "pytorch": ".pt",
        "tensorflow": ".pb",
        "keras": ".keras",
        "stable_baselines3": ".zip",
        "onnx": ".onnx",
    }
    return suffixes.get(model_type, "")


class ModelManager:
    _console: Optional[SAIConsole]
    _env: gym.Env

    _algorithm: Optional[str] = None
    _model_type: ModelLibraryType
    _model: ModelType
    _handler: BaseModelHandler
    _action_manager: Optional[ActionFunctionManager] = None

    def __init__(
        self,
        env: gym.Env,
        model: ModelType,
        model_type: Optional[ModelLibraryType] = None,
        algorithm: Optional[str] = None,
        action_function: Optional[str | Callable] = None,
        download_dir: str = config.temp_path,
        console: Optional[SAIConsole] = None,
        status: Optional[SAIStatus] = None,
    ):
        if status is not None:
            status.update("Loading model...")

        self._console = console
        self._env = env

        self._download_dir = download_dir

        self._model_type, self._algorithm = self._determine_model_type(
            model, model_type, algorithm, status
        )
        self._model, self._handler = self._load_model(model, status)

        if action_function is not None:
            self._action_manager = ActionFunctionManager(
                action_function=action_function,
                download_dir=self._download_dir,
                env=self._env,
                verbose=True,
                console=self._console,
                status=status,
            )

        self._print(status=status)

        if self._console:
            self._console.success("Successfully loaded model.")

    def _determine_model_type(
        self,
        model: ModelType,
        model_type: Optional[ModelLibraryType],
        algorithm: Optional[str],
        status: Optional[SAIStatus] = None,
    ) -> tuple[ModelLibraryType, Optional[str]]:
        determined_model_type = None
        determined_algorithm = None

        if self._console:
            self._console.debug(f"Determining model type for {model}")

        if status is not None:
            status.update("Determining model type for model...")

        if isinstance(model, str):
            if model.startswith(("http://", "https://")):
                if model_type is None:
                    raise ModelError("model_type must be provided for URL models.")
                determined_model_type = model_type
                determined_algorithm = algorithm
            elif os.path.exists(model):
                determined_model_type = self._determine_file_type(model)
                determined_algorithm = algorithm
            else:
                raise ModelError(f"Invalid model path or URL: {model}")
        else:
            model_type_detected = False
            for cls in model.__class__.__mro__:
                cls_path = f"{cls.__module__}.{cls.__name__}"
                if cls_path.startswith("torch.nn.") or "torch" in cls.__module__:
                    determined_model_type = "pytorch"
                    model_type_detected = True
                    break
                elif "tensorflow" in cls.__module__:
                    determined_model_type = "tensorflow"
                    model_type_detected = True
                    break
                elif "keras" in cls.__module__:
                    determined_model_type = "keras"
                    model_type_detected = True
                    break
                elif "stable_baselines3" in cls.__module__:
                    determined_model_type = "stable_baselines3"
                    model_type_detected = True

                    determined_algorithm = cls.__name__
                    if determined_algorithm not in [
                        "A2C",
                        "DDPG",
                        "DQN",
                        "HER",
                        "PPO",
                        "SAC",
                        "TD3",
                    ]:
                        raise ModelError(
                            f"Unsupported algorithm: {determined_algorithm}. "
                            "Supported algorithms are: A2C, DDPG, DQN, HER, PPO, SAC, TD3"
                        )
                    break
                elif "onnx" in cls.__module__:
                    determined_model_type = "onnx"
                    model_type_detected = True
                    break

            # Fallback for raw TF1 models
            sess = getattr(model, "sess", None)
            if sess:
                sess_class = sess.__class__
                if (
                    hasattr(sess_class, "__module__")
                    and "tensorflow" in sess_class.__module__.lower()
                    and "session" in sess_class.__name__.lower()
                    and hasattr(model, "states")
                ):
                    return "tensorflow", None

            if not model_type_detected:
                if model_type is not None:
                    determined_model_type = model_type
                else:
                    raise ModelError(
                        "Could not automatically determine model type. "
                        "Please specify 'model_type' parameter explicitly."
                    )

        if determined_model_type is None:
            raise ModelError(
                "Could not determine model type. "
                "Please specify 'model_type' parameter explicitly."
            )

        if model_type and determined_model_type != model_type:
            raise ModelError(
                f"Provided model_type '{model_type}' does not match "
                f"detected type '{determined_model_type}'"
            )

        if (
            determined_algorithm is None
            and determined_model_type == "stable_baselines3"
        ):
            raise ModelError(
                "Unable to determine algorithm for Stable Baselines3. "
                "Please specify 'algorithm' parameter explicitly. "
                "Supported algorithms are: A2C, DDPG, DQN, HER, PPO, SAC, TD3"
            )

        if (
            determined_algorithm is not None
            and algorithm is not None
            and determined_algorithm != algorithm
        ):
            raise ModelError(
                f"Provided algorithm '{algorithm}' does not match "
                f"detected algorithm '{determined_algorithm}'"
            )

        if self._console:
            self._console.debug(f"Determined model type: {determined_model_type}")
        return determined_model_type, determined_algorithm

    def _determine_file_type(self, file_path: str) -> ModelLibraryType:
        _, ext = os.path.splitext(file_path)
        if ext in [".pt", ".pth"]:
            return "pytorch"
        elif ext in [".pb", ".pbx"]:
            return "tensorflow"
        elif ext in [".h5", ".keras"]:
            return "keras"
        elif ext == ".zip":
            return "stable_baselines3"
        elif ext == ".onnx":
            return "onnx"
        else:
            raise ModelError(f"Unsupported file type: {ext}")

    def _load_model(self, model: ModelType, status: Optional[SAIStatus] = None):
        if self._console:
            self._console.debug(f"Loading model: {model}")

        loaded_model = None
        loaded_handler = None

        if isinstance(model, str) and model.startswith(("http://", "https://")):
            model_path = self._download_model(model, status)
            if status is not None:
                status.update("Loading model...")
                status.stop()
            loaded_handler = get_handler(
                self._env, self._model_type, self._algorithm, self._console
            )
            loaded_model = loaded_handler.load_model(model_path)

        elif self._model_type in [
            "pytorch",
            "tensorflow",
            "keras",
            "stable_baselines3",
            "onnx",
        ]:
            loaded_handler = get_handler(
                self._env, self._model_type, self._algorithm, self._console
            )
            if isinstance(model, str):
                if status is not None:
                    status.update("Loading model...")
                    status.stop()
                loaded_model = loaded_handler.load_model(model)
            else:
                loaded_model = model
        else:
            raise ModelError(f"Unsupported model type: {self._model_type}")

        if status is not None:
            status.start()

        if self._console:
            self._console.debug(
                f"Loaded model: {loaded_model} with handler: {loaded_handler}"
            )
        return loaded_model, loaded_handler

    def _download_model(self, model_url: str, status: Optional[SAIStatus] = None):
        if self._console:
            self._console.debug(f"Downloading model from {model_url}")

        if status is not None:
            status.update("Downloading model...")
            status.stop()

        os.makedirs(self._download_dir, exist_ok=True)
        filename = model_url.split("/")[-1].split("?")[0]
        file_extension = get_model_suffix(self._model_type)
        model_path = os.path.join(self._download_dir, filename + file_extension)

        try:
            if self._console:
                with self._console.progress("Downloading model") as progress:
                    with requests.get(model_url, stream=True) as response:
                        response.raise_for_status()
                        total_size = int(response.headers.get("content-length", 0))
                        task = progress.add_task(
                            "[green]Downloading...", total=total_size
                        )

                        chunk_size = 8192  # 8 KB
                        downloaded_size = 0

                        with open(model_path, "wb") as f:
                            for chunk in response.iter_content(chunk_size=chunk_size):
                                if chunk:
                                    f.write(chunk)
                                    downloaded_size += len(chunk)
                                    progress.update(task, advance=len(chunk))
            else:
                with requests.get(model_url) as response:
                    response.raise_for_status()
                    with open(model_path, "wb") as f:
                        f.write(response.content)

        except requests.exceptions.RequestException as e:
            raise NetworkError(f"Failed to download model: {e}")

        if status is not None:
            status.start()

        return model_path

    def _print(self, status: Optional[SAIStatus] = None):
        if status:
            status.update("Processing model...")

        if not self._console:
            return

        title = "Model"
        env_id = (
            self._env.spec.id
            if hasattr(self._env, "spec") and self._env.spec is not None
            else "Unknown"
        )
        info_group = f"""[bold cyan]Type:[/bold cyan]          {self._model_type}
[bold cyan]Environment:[/bold cyan]    {env_id}
[bold cyan]Action Function:[/bold cyan]  {"Custom" if self._action_manager else f"Default ({'sample' if self._handler.is_continuous else 'argmax'})"}"""

        panel_group = self._console.group(
            Align.left(Text.from_markup(info_group)),
        )

        panel = self._console.panel(panel_group, title=title, padding=(1, 2))
        self._console.print()
        self._console.print(panel)

    def get_action(self, obs):
        if self._action_manager:
            policy = self._handler.get_policy(self._model, obs)
            return self._action_manager.get_action(policy)

        return self._handler.get_action(self._model, obs)

    def save_model(self, model_path, use_onnx=False, **kwargs):
        if os.path.exists(model_path):
            raise ModelError(f"Model path already exists: {model_path}")

        os.makedirs(os.path.dirname(model_path), exist_ok=True)

        if use_onnx:
            self._handler.export_to_onnx(self._model, model_path)
        else:
            self._handler.save_model(self._model, model_path)

    def save_action_function(self, action_fn_path: str):
        if self._action_manager:
            self._action_manager.save_action_function(action_fn_path)
        else:
            raise ModelError("No action function loaded.")

    def hash(self):
        if self._model_type not in SUPPORTED_HASHING_MODEL_TYPES:
            return {}

        return self._handler.hash(self._model)

    @property
    def algorithm(self):
        return self._algorithm

    @property
    def model_type(self):
        return self._model_type

    @property
    def model(self):
        return self._model



class TeeIO:
    def __init__(self, *files):
        self.files = files

    def write(self, data):
        for f in self.files:
            f.write(data)
            f.flush()

    def flush(self):
        for f in self.files:
            f.flush()



from typing import Optional, Any

from sai_rl.model.handlers._base import BaseModelHandler
from sai_rl.sai_console import SAIConsole

import gymnasium as gym
import numpy as np
import tarfile
import tempfile
import os

IMPORT_ERROR_MESSAGE = (
    "TensorFlow is not installed. Please install it using 'pip install sai_rl[tf]'."
)


class TensorFlowModelHandler(BaseModelHandler):
    _tf: Any = None
    _tf_v1: Any = None
    _tf2onnx: Any = None
    _onnx: Any = None
    _tf_available: bool = False

    def __init__(self, env: gym.Env, console: Optional[SAIConsole] = None):
        super().__init__(env, console)
        try:
            import tensorflow.compat.v1 as tf_v1
            import tensorflow as tf
            import tf2onnx
            import onnx

            self._tf = tf
            self._tf_v1 = tf_v1
            self._tf2onnx = tf2onnx
            self._onnx = onnx
            self._tf_available = True
        except ImportError:
            self._tf_available = False
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def _check_tensorflow_available(self) -> None:
        if not self._tf_available:
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def is_tf1_model(self, model) -> bool:
        return hasattr(model, "sess")

    def is_tf2_model(self, model) -> bool:
        return (callable(model) and hasattr(model, "__call__")) or isinstance(
            model, self._tf.Module
        )

    def load_model(self, model_path: str):
        self._check_tensorflow_available()

        model = None
        tf2_load_successful = False
        try:
            model = self._load_model_v2(model_path)
            tf2_load_successful = self.is_tf2_model(model)
        except Exception as e:
            print(f"TF2 model load failed: {str(e)}")

        if tf2_load_successful:
            return model
        elif model is not None:
            raise ValueError("Loaded TF2 model is not recognized as valid.")

        print("Falling back to TF1 model loader...")
        return self._load_model_v1(model_path)

    def _load_model_v2(self, model_path):
        self._check_tensorflow_available()

        class TF2ModelWrapper:
            def __init__(self, model_path, tf):
                temp_dir = tempfile.mkdtemp()
                with tarfile.open(model_path, "r") as tar:
                    tar.extractall(temp_dir)

                model = tf.saved_model.load(temp_dir)
                self.infer = model.signatures["serving_default"]
                self.output_key = list(self.infer.structured_outputs.keys())[0]

            @self._tf.function
            def __call__(self, x):
                if x.ndim == 1:
                    x = np.expand_dims(x, axis=0)
                state = self._tf.convert_to_tensor(x, dtype=self._tf.float32)
                return self.infer(state)[self.output_key].numpy()

        return TF2ModelWrapper(model_path, self._tf)

    def _load_model_v1(self, model_path):
        self._check_tensorflow_available()

        self._tf_v1.disable_eager_execution()

        class FrozenTF1Model:
            def __init__(
                self,
                model_path: str,
                input_name: str = "states:0",
                output_name: str = "policy:0",
                tf = None
            ):
                self.graph = tf.Graph()
                with tf.gfile.GFile(model_path, "rb") as f:
                    graph_def = tf.GraphDef()
                    graph_def.ParseFromString(f.read())

                with self.graph.as_default():
                    tf.import_graph_def(graph_def, name="")

                self.sess = tf.Session(graph=self.graph)
                self.states = self.graph.get_tensor_by_name(input_name)
                self.policy = self.graph.get_tensor_by_name(output_name)

            def select_action(self, state: np.ndarray):
                state = np.expand_dims(state, axis=0) if state.ndim == 1 else state
                return self.sess.run(self.policy, feed_dict={self.states: state})[0]

        return FrozenTF1Model(model_path, input_name="states:0", output_name="policy:0", tf=self._tf_v1)

    def _get_parameters_array(self, model):
        if self.is_tf1_model(model):
            values = model.sess.run(self._tf_v1.trainable_variables())
            return [v.tolist() for v in values]
        elif self.is_tf2_model(model):
            return [var.numpy().tolist() for var in model.variables]

    def _infer_architecture(self, model):
        if self.is_tf1_model(model):
            return self._infer_architecture_v1(model)
        elif self.is_tf2_model(model):
            return self._infer_architecture_v2(model)

    def _infer_architecture_v2(self, model):
        structure = []
        for var in model.variables:
            var_info = {
                "name": var.name,
                "shape": tuple(var.shape.as_list()),
                "dtype": str(var.dtype),
                "trainable": var.trainable,
            }
            structure.append(var_info)
        return structure

    def _infer_architecture_v1(self, model):
        graph = model.sess.graph
        graph_def = graph.as_graph_def()

        layers = []
        for node in graph_def.node:
            if node.op in {
                "MatMul",
                "Conv2D",
                "BiasAdd",
                "Relu",
                "Tanh",
                "Sigmoid",
                "Elu",
                "Softmax",
            }:
                layers.append(
                    {
                        "name": node.name,
                        "op": node.op,
                        "inputs": node.input[:],  # list of input tensor names
                    }
                )

        variables = []
        for var in self._tf_v1.trainable_variables():
            variables.append(
                {
                    "name": var.name,
                    "shape": tuple(var.shape.as_list()),
                    "dtype": var.dtype.name,
                }
            )

        return {
            "class": model.__class__.__name__,
            "layers": layers,
            "variables": variables,
        }

    def get_policy(self, model, obs: np.ndarray) -> np.ndarray:
        self._check_tensorflow_available()

        if len(obs.shape) < self.obs_space_dims + 1:
            obs = np.expand_dims(obs, axis=0)

        if self.is_tf1_model(model):
            return model.sess.run(model.policy, feed_dict={model.states: obs})
        elif self.is_tf2_model(model):
            return np.array(model(obs))
        elif hasattr(model, "predict"):
            return np.array(model.predict(obs, verbose=0))
        else:
            raise ValueError("Unknown model type: cannot compute policy.")

    @staticmethod
    def _save_as_pbx(source_dir, pbx_path):
        with tarfile.open(pbx_path, "w") as tar:
            tar.add(source_dir, arcname="")

    def save_model(self, model, model_path):
        self._check_tensorflow_available()

        if self.is_tf1_model(model):
            from tensorflow.compat.v1.graph_util import convert_variables_to_constants

            sess = model.sess
            output_node_names = [model.policy.name.split(":")[0]]
            frozen_graph_def = convert_variables_to_constants(
                sess, sess.graph.as_graph_def(), output_node_names=output_node_names
            )
            self._tf_v1.io.write_graph(
                frozen_graph_def, logdir=".", name=model_path, as_text=False
            )
        elif self.is_tf2_model(model):
            folder_name = model_path.replace(".pb", "")
            concrete_fn = model.get_concrete_function()
            self._tf.saved_model.save(model, folder_name, signatures={"serving_default": concrete_fn})
            self._save_as_pbx(folder_name, model_path + ".pbx")

        return model_path

    def export_to_onnx_v1(self, model, model_path: str):
        from tensorflow.compat.v1.graph_util import convert_variables_to_constants

        sess = model.sess
        input_names = [model.states.name]
        output_names = [model.policy.name]

        # Freeze variables into constants
        frozen_graph_def = convert_variables_to_constants(
            sess,
            sess.graph_def,
            output_node_names=[name.split(":")[0] for name in output_names],
        )

        onnx_model, _ = self._tf2onnx.convert.from_graph_def(
            frozen_graph_def,
            input_names=input_names,
            output_names=output_names,
            opset=11,
        )

        self._onnx.save_model(
            onnx_model,
            model_path if model_path.endswith(".onnx") else f"{model_path}.onnx",
        )

    def export_to_onnx_v2(self, model, model_path: str):
        input_shape = [None, *self.observation_space.shape]
        input_spec = self._tf.TensorSpec(input_shape, self._tf.float32, name="input")

        # Wrap model in a tracing function
        @self._tf.function
        def wrapped_model(x):
            return model(x)

        # Manually create concrete function
        concrete_func = wrapped_model.get_concrete_function(input_spec)

        onnx_model, _ = self._tf2onnx.convert.from_function(
            wrapped_model, input_signature=[input_spec], opset=11
        )

        self._onnx.save_model(
            onnx_model,
            model_path if model_path.endswith(".onnx") else f"{model_path}.onnx",
        )

    def hash(self, model):
        return {
            "architecture_hash": self._hash(self._infer_architecture(model)),
            "parameters_hash": self._hash(self._get_parameters_array(model)),
        }

    def export_to_onnx(self, model, model_path: str):
        self._check_tensorflow_available()
        if self.is_tf1_model(model):
            self.export_to_onnx_v1(model, model_path)
        elif self.is_tf2_model(model):
            self.export_to_onnx_v2(model, model_path)



from typing import Optional, Union

import numpy as np
import time

from rich.table import Table
from rich.panel import Panel
from rich.align import Align

from sai_rl.sai_console import SAIConsole


def generate_table(
    console: SAIConsole,
    timestep: int = 0,
    active_count: int = 0,
    num_envs: int = 0,
    current_score: Union[float, np.float64] = 0,
    current_fps: float = 0,
    time_elapsed: float = 0,
    timeout: Optional[int] = None,
) -> Table:
    table = console.table(title=None, show_header=True, show_lines=False)
    table.add_column("Steps", justify="center", style="cyan")
    table.add_column("Finished Envs", justify="center", style="green")
    table.add_column("Completed", justify="center", style="magenta")
    table.add_column("Avg Score", justify="center", style="green")
    table.add_column("FPS", justify="center", style="blue")
    table.add_column("Time", justify="center", style="yellow")
    table.add_column("Remaining", justify="center", style="red")

    completed = num_envs - active_count
    completion_percentage = (
        f"{(completed / num_envs) * 100:.1f}%" if num_envs > 0 else "0%"
    )

    table.add_row(
        str(timestep),
        f"{completed}/{num_envs}",
        completion_percentage,
        f"{current_score:.2f}",
        f"{current_fps:.1f}",
        f"{time_elapsed:.1f}s",
        f"{timeout - time_elapsed:.1f}s" if timeout else "0.0s",
    )

    return table


def generate_batch_panel(
    console: SAIConsole,
    batch_active_count: int = 0,
    batch_size: int = 0,
    batch_idx: int = 0,
    batch_start_time: float = 0,
    batch_score: float = 0,
    batch_timesteps: int = 0,
    batch_timeout: Optional[int] = None,
    num_envs: int = 0,
    num_batches: int = 0,
    is_final: bool = False,
) -> Panel:
    now = time.time() - batch_start_time
    fps_now = batch_timesteps / now if now > 0 else 0
    table = generate_table(
        console,
        batch_timesteps,
        batch_active_count if not is_final else 0,
        batch_size,
        batch_score,
        fps_now,
        now,
        batch_timeout if not is_final else 0,
    )
    return console.panel(
        Align.center(table),
        title=f"Benchmark Progress (Batch {batch_idx + 1}/{num_batches})",
        subtitle=f"{(batch_size - (batch_active_count if not is_final else 0) + batch_idx * batch_size)}/{num_envs} envs",
        border_style="blue",
        title_align="center",
    )



from typing import Callable, Optional, Any
import math
import multiprocessing as mp
import gymnasium as gym
import numpy as np
import time
import platform

from rich.live import Live

from sai_rl.model import ModelManager
from sai_rl.sai_console import SAIConsole, SAIStatus
from sai_rl.benchmark.types import BenchmarkResults
from sai_rl.benchmark.utils import generate_batch_panel
from sai_rl.benchmark.seed import get_match_seed


# TODO: make this simpler and cleaner
def run_benchmark(
    env_creator: Callable[[], gym.Env],
    num_envs: int = 1,
    seed: Optional[int] = None,
    model: Optional[ModelManager] = None,
    timeout: int = 600,
    console: Optional[SAIConsole] = None,
    status: Optional[SAIStatus] = None,
    show_progress: bool = False,
    save_actions: bool = False,
) -> tuple[BenchmarkResults, list[Any], Optional[list[int]]]:
    """
    Run a benchmark on the given environment and model in CPU‐sized batches,
    reusing a single Live display for the entire run.
    Always returns (results, actions, seeds), even on error.
    """
    start_time = time.time()

    if status:
        status.update("Starting benchmark...")
        status.stop()

    results: BenchmarkResults = {
        "status": "error",
        "score": None,
        "duration": 0,
        "matches": [],
        "logs": None,
        "error": None,
    }

    all_actions: list[list[Any]] = [] if save_actions else []
    all_seeds: list[int] = []

    try:
        is_windows = platform.system() == "Windows"
        if is_windows:
            mp.set_start_method("spawn", force=True)
        else:
            mp.set_start_method("fork", force=True)

        # Calculate Batch Size for Parallel Processing
        cpu_count = mp.cpu_count()
        batch_size = min(num_envs, cpu_count)
        num_batches = math.ceil(num_envs / batch_size)

        # Using the Competition Seed or a Random One
        # -> this lets us reproduce the environment for when we want to replay the benchmark
        #    for recording specific episodes after the benchmark is done.
        master_seed = seed if seed is not None else np.random.randint(0, 2**31 - 1)

        # Aggregate Results Across Batches
        agg_scores: list[float] = []
        agg_durations: list[float] = []
        agg_timesteps: list[int] = []

        reached_timeout = False

        # Create the Benchmark Tracker
        batch_live: Optional[Live] = None
        if show_progress and console:
            batch_live = Live(
                generate_batch_panel(console),
                console=console.console,
                refresh_per_second=15,
            )
            batch_live.start()

        for batch_idx in range(num_batches):
            batch_start_time = time.time()
            current_batch_size = min(batch_size, num_envs - batch_idx * batch_size)

            if console and not show_progress:
                console.info(
                    f"→ Batch {batch_idx + 1}/{num_batches}: {current_batch_size} envs"
                )

            # Append Current Batch Seeds
            batch_seeds = [
                get_match_seed(master_seed, batch_idx * batch_size + i)
                for i in range(current_batch_size)
            ]
            all_seeds.extend(batch_seeds)

            # Create the Environments
            # -> We found issues with the AsyncVectorEnv on Windows, so we use SyncVectorEnv instead
            #    for now. We should revisit this later to improve performance.
            envs: Optional[gym.vector.VectorEnv] = None
            if is_windows:
                envs = gym.vector.SyncVectorEnv(
                    [env_creator for _ in range(current_batch_size)],
                    copy=False,
                )
            else:
                envs = gym.vector.AsyncVectorEnv(
                    [env_creator for _ in range(current_batch_size)],
                    shared_memory=True,
                    copy=False,
                )

            if envs is None:
                raise RuntimeError("Failed to create environment")

            # Figure out the Environment Timeout
            # -> this is to make sure the timeout is the timeout of the time inside the environment
            #    and not the time it takes to run the benchmark. On slower machines, this can lead to
            #    early timeouts if the environment is slow.
            fps: int = envs.metadata.get("render_fps", 30)
            timeout_timestep = timeout * fps

            # Create the State to Track the Environment
            alive = np.ones(current_batch_size, bool)
            terminated = np.zeros(current_batch_size, bool)
            truncated = np.zeros(current_batch_size, bool)
            scores = np.zeros(current_batch_size, float)
            timesteps = np.zeros(current_batch_size, int)
            batch_actions = (
                [[] for _ in range(current_batch_size)] if save_actions else None
            )

            def update_status(is_final: bool = False) -> None:
                if batch_live and console:
                    batch_live.update(
                        generate_batch_panel(
                            console,
                            batch_active_count=int(alive.sum()),
                            batch_size=current_batch_size,
                            batch_idx=batch_idx,
                            batch_start_time=batch_start_time,
                            batch_score=float(scores.mean()),
                            batch_timesteps=timesteps.sum(),
                            batch_timeout=timeout,
                            num_envs=num_envs,
                            num_batches=num_batches,
                            is_final=is_final,
                        )
                    )

            # Reset the Environments
            timestep = 0
            obs, _ = envs.reset(seed=batch_seeds)
            update_status()

            # Batch Environment Loop
            while not np.all(terminated | truncated):
                if timestep > timeout_timestep:
                    if console:
                        console.warning(
                            f"Hit timeout at {timestep / fps:.1f}s; stopping current batch."
                        )
                    reached_timeout = True
                    break

                actions = model.get_action(obs) if model else envs.action_space.sample()
                obs, rewards, terms, truns, _ = envs.step(actions)

                terminated |= terms
                truncated |= truns
                alive = ~(terminated | truncated)
                scores += rewards * alive
                timesteps += alive.astype(int)
                timestep += 1

                if save_actions:
                    for i in range(current_batch_size):
                        if alive[i]:
                            batch_actions[i].append(actions[i])  # type: ignore

                update_status()

            update_status(True)

            # Save the Batch Results
            if save_actions and batch_actions is not None:
                all_actions.extend(batch_actions)

            agg_scores.extend(scores.tolist())
            agg_durations.extend((timesteps / fps).tolist())
            agg_timesteps.extend(timesteps.tolist())

            try:
                envs.close()
            except:  # noqa: E722
                pass

        # Calculate the Final Results
        duration = time.time() - start_time
        avg_score = float(np.mean(agg_scores)) if agg_scores else 0.0

        results["status"] = "success" if not reached_timeout else "timeout"
        results["score"] = avg_score  # type: ignore
        results["duration"] = duration
        results["matches"] = [
            {
                "score": agg_scores[i],
                "duration": agg_durations[i],
            }
            for i in range(num_envs)
        ]
        results["logs"] = None
        results["error"] = None

        if batch_live:
            batch_live.stop()

        if console:
            console.info(
                f"\n\nBenchmark of {num_envs} episodes completed in {results['duration']:.2f} seconds"
            )
            console.info(f"Average Episode Score: {results['score']:.2f}")
            console.info(
                f"Average Duration per Batch: {np.mean(agg_durations):.2f} seconds"
            )
            console.info(
                f"Average FPS Across All Batches: {np.sum(agg_timesteps) / results['duration']:.2f}"
            )
            console.info(f"Total Timesteps: {np.sum(agg_timesteps)}\n")

    except Exception as e:
        results["status"] = "error"
        results["error"] = str(e)
        if console:
            console.error(f"Error during benchmark: {e!r}")

    finally:
        if batch_live:
            batch_live.stop()
        if status:
            status.update("Cleaning up...")

    return results, all_actions, all_seeds



from typing import Any, Optional
from sai_rl.model.handlers._base import BaseModelHandler

from sai_rl.sai_console import SAIConsole

import numpy as np
import gymnasium as gym

IMPORT_ERROR_MESSAGE = (
    "ONNX Runtime is not installed. Please install it using 'pip install onnxruntime'."
)


class OnnxModelHandler(BaseModelHandler):
    _ort: Any = None
    _ort_available: bool = False

    def __init__(self, env: gym.Env, console: Optional[SAIConsole] = None):
        super().__init__(env, console)
        try:
            import onnxruntime as ort

            self._ort = ort
            self._ort_available = True
        except ImportError:
            self._ort_available = False
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def _check_ort_available(self) -> None:
        if not self._ort_available:
            raise ImportError(IMPORT_ERROR_MESSAGE)

    def load_model(self, model_path: str):
        self._check_ort_available()
        return self._ort.InferenceSession(
            model_path, providers=["CPUExecutionProvider"]
        )

    def get_policy(self, model: Any, obs: np.ndarray):
        """
        Get the action from the model given an observation.
        Args:
            model (onnxruntime.InferenceSession): The ONNX model.
            obs (np.ndarray): The observation input to the model.
            is_continuous (bool): If True, the model is for continuous action space.
        Returns:
            int: The action selected by the model.
        Raises:
            ImportError: If PyTorch is not installed.
        """

        self._check_ort_available()

        if len(obs.shape) < self.obs_space_dims + 1:
            obs = np.expand_dims(obs, axis=0)

        return model.run(None, {"input": obs})[0]

    def save_model(self, model, model_path):
        self._check_ort_available()
        if self._console:
            self._console.warning(
                "Saving ONNX models is not supported. The model will not be saved."
            )
        pass

    def export_to_onnx(self, model, model_path):
        self._check_ort_available()
        if self._console:
            self._console.warning(
                "Exporting to ONNX is not supported. The model will not be exported."
            )
        pass

    def _onnx_type_to_numpy_type(self, onnx_type):
        """Convert ONNX data type to numpy data type."""
        # ONNX type mapping to numpy types
        type_map = {
            1: np.float32,  # FLOAT
            2: np.uint8,  # UINT8
            3: np.int8,  # INT8
            4: np.uint16,  # UINT16
            5: np.int16,  # INT16
            6: np.int32,  # INT32
            7: np.int64,  # INT64
            8: str,  # STRING
            9: bool,  # BOOL
            10: np.float16,  # FLOAT16
            11: np.float64,  # DOUBLE
            12: np.uint32,  # UINT32
            13: np.uint64,  # UINT64
        }
        return type_map.get(onnx_type, np.float32)

    def _get_parameters_array(self, model):
        self._check_ort_available()
        if hasattr(model, "_model_path"):
            import onnx

            onnx_model = onnx.load(model._model_path)
            model_proto = onnx_model
        else:
            model_proto = None

        parameters = []

        if model_proto is not None:
            for initializer in model_proto.graph.initializer:
                if initializer.raw_data:
                    tensor_data = np.frombuffer(
                        initializer.raw_data,
                        dtype=self._onnx_type_to_numpy_type(initializer.data_type),
                    )

                    if initializer.dims:
                        tensor_data = tensor_data.reshape(initializer.dims)

                    parameters.append(tensor_data.tolist())
                else:
                    if hasattr(initializer, "float_data") and initializer.float_data:
                        parameters.append(list(initializer.float_data))
                    elif hasattr(initializer, "int64_data") and initializer.int64_data:
                        parameters.append(list(initializer.int64_data))
                    elif hasattr(initializer, "int32_data") and initializer.int32_data:
                        parameters.append(list(initializer.int32_data))
        return parameters

    def _infer_architecture(self, model):
        self._check_ort_available()

        model_meta = model.get_modelmeta()
        inputs_meta = model.get_inputs()
        outputs_meta = model.get_outputs()

        architecture = {"inputs": [], "outputs": []}

        for input_info in inputs_meta:
            input_data = {
                "name": input_info.name,
                "shape": list(input_info.shape)
                if hasattr(input_info, "shape")
                else None,
                "type": str(input_info.type) if hasattr(input_info, "type") else None,
            }
            architecture["inputs"].append(input_data)

        for output_info in outputs_meta:
            output_data = {
                "name": output_info.name,
                "shape": list(output_info.shape)
                if hasattr(output_info, "shape")
                else None,
                "type": str(output_info.type) if hasattr(output_info, "type") else None,
            }
            architecture["outputs"].append(output_data)

        # Add action and observation space info
        architecture["spaces"] = {
            "observation_space": {
                "shape": list(self.observation_space.shape),
                "dtype": str(self.observation_space.dtype),
            },
            "action_space": {
                "shape": list(self.action_space.shape)
                if hasattr(self.action_space, "shape")
                else None,
                "dtype": str(self.action_space.dtype)
                if hasattr(self.action_space, "dtype")
                else None,
                "is_continuous": self.is_continuous,
            },
        }

        print(architecture)

        return architecture

    def hash(self, model):
        self._check_ort_available()

        return {
            "architecture_hash": self._hash(self._infer_architecture(model)),
            "parameters_hash": self._hash(self._get_parameters_array(model)),
        }



from typing import Any, List, Mapping, Optional, TypedDict
from types import SimpleNamespace
import json

from sai_rl.api.routes.environment import EnvironmentType
from sai_rl.api.requestor import APIRequestor


class CompetitionType(TypedDict):
    id: str
    slug: str
    name: str
    opensource: bool
    environment: EnvironmentType
    environmentVariables: Mapping[str, Any]
    numberOfBenchmarks: Optional[int]
    seed: Optional[int]


class CompetitionAPI:
    def __init__(self, api: APIRequestor):
        self._api = api

    def get(self, competition_id: str) -> Optional[CompetitionType]:
        response = self._api.get(f"/v1/competitions/{competition_id}")

        if not response:
            return None

        raw_competition = response.json()
        if not raw_competition:
            return None

        comp_env = json.loads(
            json.dumps(raw_competition), object_hook=lambda d: SimpleNamespace(**d)
        )
        competition: CompetitionType = {
            "id": comp_env.id,
            "slug": comp_env.slug,
            "name": comp_env.name,
            "opensource": comp_env.opensource if hasattr(comp_env, "opensource") else False,
            "environment": {
                "id": comp_env.environment.id,
                "name": comp_env.environment.name,
                "gymId": comp_env.environment.gymId,
                "type": comp_env.environment.type,
                "actionType": comp_env.environment.actionType,
                "package": {
                    "id": comp_env.environment.package.id,
                    "name": comp_env.environment.package.name,
                    "version": comp_env.environment.package.version,
                },
            },
            "environmentVariables": vars(comp_env.environmentVariables)
            if hasattr(comp_env, "environmentVariables")
            and not isinstance(comp_env.environmentVariables, dict)
            else (
                comp_env.environmentVariables
                if hasattr(comp_env, "environmentVariables")
                else {}
            ),
            "numberOfBenchmarks": comp_env.numberOfBenchmarks
            if hasattr(comp_env, "numberOfBenchmarks")
            else None,
            "seed": comp_env.seed if hasattr(comp_env, "seed") else None,
        }

        return competition

    def list(self) -> Optional[List[CompetitionType]]:
        response = self._api.get("/v1/competitions")

        if not response:
            return None

        raw_competitions = response.json()
        if not raw_competitions:
            return None

        competitions: List[CompetitionType] = []

        for raw_competition in raw_competitions:
            comp_env = json.loads(
                json.dumps(raw_competition), object_hook=lambda d: SimpleNamespace(**d)
            )
            competition: CompetitionType = {
                "id": comp_env.id,
                "slug": comp_env.slug,
                "name": comp_env.name,
                "opensource": comp_env.opensource if hasattr(comp_env, "opensource") else False,
                "environment": {
                    "id": comp_env.environment.id,
                    "name": comp_env.environment.name,
                    "gymId": comp_env.environment.gymnasiumEnv,
                    "type": comp_env.environment.type,
                    "actionType": comp_env.environment.actionType,
                    "package": {
                        "id": comp_env.environment.package.id,
                        "name": comp_env.environment.package.name,
                        "version": comp_env.environment.package.version,
                    },
                },
                "environmentVariables": comp_env.environmentVariables,
                "numberOfBenchmarks": comp_env.numberOfBenchmarks
                if hasattr(comp_env, "numberOfBenchmarks")
                else None,
                "seed": comp_env.seed if hasattr(comp_env, "seed") else None,
            }

            competitions.append(competition)

        return competitions



import hashlib
import hmac
import struct


def get_match_seed(master_int: int, index: int) -> int:
    master_key = master_int.to_bytes(32, "big")
    prk = hmac.new(b"\x00" * 32, master_key, hashlib.sha256).digest()
    info = struct.pack(">Q", index)
    t = b""
    okm = b""
    counter = 1
    while len(okm) < 8:
        t = hmac.new(prk, t + info + bytes([counter]), hashlib.sha256).digest()
        okm += t
        counter += 1
    return int.from_bytes(okm[:8], "big")



from typing import Literal, Any

EnvironmentActionType = Literal["discrete", "continuous"]
EnvironmentStandardType = Literal[
    "gymnasium", "gym-v26", "gym-v21", "pettingzoo", "pufferlib"
]

ModelType = Any
ModelLibraryType = Literal["pytorch", "tensorflow", "keras", "stable_baselines3", "onnx"]



from typing import Any, TYPE_CHECKING, Optional
from sai_rl.model.handlers._base import BaseModelHandler

from sai_rl.sai_console import SAIConsole

import gymnasium as gym
import numpy as np

if TYPE_CHECKING:
    from stable_baselines3.common.policies import BasePolicy

    SB3BasePolicy = BasePolicy
else:
    SB3BasePolicy = Any

IMPORT_ERROR_MESSAGE = (
    "Stable Baselines3 is not installed. Please install it using 'pip install stable-baselines3'."
    " If you are using a custom model, please ensure it is compatible with Stable Baselines3."
)

SB3_ALGORITHMS = ["A2C", "DDPG", "DQN", "HER", "PPO", "SAC", "TD3"]


class SBL3ModelHandler(BaseModelHandler):
    _torch: Any = None
    _sb3: Any = None
    _sb3_available: bool = False

    def __init__(self, env: gym.Env, console: Optional[SAIConsole] = None, algo="PPO"):
        super().__init__(env, console)

        try:
            import stable_baselines3 as sb3
            import torch

            self._sb3 = sb3
            self._torch = torch
            self._sb3_available = True
        except ImportError:
            self._sb3_available = False
            raise ImportError(IMPORT_ERROR_MESSAGE)

        algo = algo.upper()
        if algo not in SB3_ALGORITHMS:
            raise ValueError(
                f"Algorithm {algo} is not supported. Please use one of the following: {SB3_ALGORITHMS}"
            )

        self._algo = getattr(sb3, algo)

        class SB3TorchPolicy(torch.nn.Module):
            def __init__(self, policy: SB3BasePolicy):
                super().__init__()
                self.policy = policy

            def forward(self, observation: torch.Tensor):
                return self.policy(observation, deterministic=True)

        self._policy_class = SB3TorchPolicy

    def _check_sb3_available(self) -> None:
        if not self._sb3_available:
            raise ImportError(
                "Stable Baselines3 is not installed. Please install it using 'pip install stable-baselines3'."
            )

    def load_model(self, model_path) -> SB3BasePolicy:
        try:
            return self._algo.load(model_path)
        except Exception as e:
            raise ValueError(
                f"Error loading model: {e}. We do only support PPO models from Stable Baselines3 at the moment. We are working on supporting other models."
            ) from e

    def get_policy(self, model: SB3BasePolicy, obs: Any) -> Any:
        if len(obs.shape) < self.obs_space_dims + 1:
            obs = np.expand_dims(obs, axis=0)

        return model.predict(obs, deterministic=True)[0]

    def get_action(self, model, obs: np.ndarray) -> np.ndarray:
        return self.get_policy(model, obs)

    def save_model(self, model, model_path):
        self._check_sb3_available()

        model.save(model_path)
        return model_path

    def export_to_onnx(self, model: SB3BasePolicy, model_path: str):
        self._check_sb3_available()

        torch_model = self._policy_class(model.policy)
        torch_model.eval()

        # Convert numpy dtype to torch dtype
        numpy_to_torch_dtype = {
            'float32': self._torch.float32,
            'float64': self._torch.float64,
            'int32': self._torch.int32,
            'int64': self._torch.int64,
        }
        
        obs_dtype_str = str(self.observation_space.dtype)
        torch_dtype = numpy_to_torch_dtype.get(obs_dtype_str, self._torch.float32)

        torch_input = self._torch.randn(
            1,
            *self.observation_space.shape,
            dtype=torch_dtype,
        )

        return self._torch.onnx.export(
            torch_model,
            (torch_input,),
            model_path,
            export_params=True,
            opset_version=18,
            input_names=["input"],
            output_names=["output"],
            dynamic_axes={"input": {0: "batch_size"}, "output": {0: "batch_size"}},
        )

    def _get_parameters_array(self, model: SB3BasePolicy):
        self._check_sb3_available()

        policy_params = []

        if hasattr(model, "policy") and hasattr(model.policy, "parameters"):
            for param in model.policy.parameters():
                policy_params.append(param.detach().cpu().numpy().tolist())

        if hasattr(model, "value_net") and hasattr(model.value_net, "parameters"):
            for param in model.value_net.parameters():
                policy_params.append(param.detach().cpu().numpy().tolist())

        if hasattr(model, "get_parameters"):
            params_dict = model.get_parameters()
            if "policy" in params_dict:
                policy_state = params_dict["policy"]
                for key, value in policy_state.items():
                    if hasattr(value, "detach"):
                        policy_params.append(value.detach().cpu().numpy().tolist())
                    elif isinstance(value, np.ndarray):
                        policy_params.append(value.tolist())

        return policy_params

    def _infer_architecture(self, model: SB3BasePolicy):
        self._check_sb3_available()

        architecture = {
            "class": model.__class__.__name__,
            "algorithm": getattr(model, "_get_name", lambda: "Unknown")(),
            "policy_class": model.policy.__class__.__name__
            if hasattr(model, "policy")
            else None,
            "layers": [],
            "network_architecture": {},
        }

        if hasattr(model, "policy"):
            policy = model.policy

            if hasattr(policy, "mlp_extractor"):
                # For on-policy algorithms like PPO, A2C
                mlp_extractor = policy.mlp_extractor
                if hasattr(mlp_extractor, "policy_net"):
                    for i, layer in enumerate(mlp_extractor.policy_net):
                        if hasattr(layer, "weight"):
                            architecture["layers"].append(
                                {
                                    "name": f"policy_layer_{i}",
                                    "type": layer.__class__.__name__,
                                    "input_features": layer.in_features
                                    if hasattr(layer, "in_features")
                                    else None,
                                    "output_features": layer.out_features
                                    if hasattr(layer, "out_features")
                                    else None,
                                    "activation": getattr(
                                        layer, "activation", None
                                    ).__class__.__name__
                                    if hasattr(layer, "activation")
                                    else None,
                                }
                            )

                if hasattr(mlp_extractor, "value_net"):
                    for i, layer in enumerate(mlp_extractor.value_net):
                        if hasattr(layer, "weight"):
                            architecture["layers"].append(
                                {
                                    "name": f"value_layer_{i}",
                                    "type": layer.__class__.__name__,
                                    "input_features": layer.in_features
                                    if hasattr(layer, "in_features")
                                    else None,
                                    "output_features": layer.out_features
                                    if hasattr(layer, "out_features")
                                    else None,
                                    "activation": getattr(
                                        layer, "activation", None
                                    ).__class__.__name__
                                    if hasattr(layer, "activation")
                                    else None,
                                }
                            )

            elif hasattr(policy, "q_net"):
                # For off-policy algorithms like DQN, SAC
                q_net = policy.q_net
                for i, layer in enumerate(q_net):
                    if hasattr(layer, "weight"):
                        architecture["layers"].append(
                            {
                                "name": f"q_layer_{i}",
                                "type": layer.__class__.__name__,
                                "input_features": layer.in_features
                                if hasattr(layer, "in_features")
                                else None,
                                "output_features": layer.out_features
                                if hasattr(layer, "out_features")
                                else None,
                            }
                        )

            if hasattr(policy, "net_arch"):
                architecture["network_architecture"]["net_arch"] = policy.net_arch

        if hasattr(model, "learning_rate"):
            architecture["hyperparameters"] = {
                "learning_rate": float(model.learning_rate)
                if callable(model.learning_rate)
                else model.learning_rate,
            }

        architecture["spaces"] = {
            "observation_space": {
                "shape": self.observation_space.shape,
                "dtype": str(self.observation_space.dtype),
            },
            "action_space": {
                "shape": self.action_space.shape
                if hasattr(self.action_space, "shape")
                else None,
                "dtype": str(self.action_space.dtype)
                if hasattr(self.action_space, "dtype")
                else None,
                "is_continuous": self.is_continuous,
            },
        }

        return architecture

    def hash(self, model: SB3BasePolicy):
        self._check_sb3_available()

        return {
            "architecture_hash": self._hash(self._infer_architecture(model)),
            "parameters_hash": self._hash(self._get_parameters_array(model)),
        }



from typing import Callable, Optional
import gymnasium as gym
import imageio
import os
import subprocess
import time
from imageio_ffmpeg import get_ffmpeg_exe


class RecordingError(Exception):
    pass


def record_episode(
    env_creator: Callable[[str], gym.Env],
    render_mode: str,
    seed: int,
    actions: list[int],
    output_path: str,
    use_virtual_display: bool = False,
    max_steps: Optional[int] = None,
    ffmpeg_capture_size: str = "1280x720",
) -> None:
    env = env_creator(render_mode=render_mode)  # type: ignore
    env.reset(seed=seed)

    fps = env.metadata.get("render_fps", 60)
    render_mode = env.metadata.get("render_mode")

    os.makedirs(
        os.path.dirname(output_path) or ".",
        exist_ok=True,
    )

    writer = None
    ffmpeg_process = None
    original_sdl = os.environ.get("SDL_VIDEODRIVER")

    if use_virtual_display:
        os.environ["SDL_VIDEODRIVER"] = "dummy"

    try:
        if render_mode == "human":
            if "DISPLAY" not in os.environ:
                raise RecordingError("No DISPLAY for human mode recording.")
            ffmpeg_exe = get_ffmpeg_exe()
            cmd = [
                ffmpeg_exe,
                "-f",
                "x11grab",
                "-s",
                ffmpeg_capture_size,
                "-r",
                str(fps),
                "-i",
                os.environ["DISPLAY"],
                "-c:v",
                "libx264",
                "-preset",
                "ultrafast",
                "-pix_fmt",
                "yuv420p",
                "-y",
                output_path,
            ]
            ffmpeg_process = subprocess.Popen(
                cmd,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                text=True,
                bufsize=1,
                universal_newlines=True,
            )
            time.sleep(1.5)
            if ffmpeg_process.poll() is not None:
                out, err = ffmpeg_process.communicate()
                raise RecordingError(
                    f"FFmpeg error {ffmpeg_process.returncode}\n{out}\n{err}"
                )
        else:
            writer = imageio.get_writer(output_path, fps=fps)

        for i, action in enumerate(actions):
            if max_steps is not None and i >= max_steps:
                break

            _, _, terminated, truncated, _ = env.step(action)
            frame = env.render()
            if writer and frame is not None:
                writer.append_data(frame)
            if terminated or truncated:
                break

    except Exception:
        if ffmpeg_process and ffmpeg_process.poll() is None:
            ffmpeg_process.terminate()
            try:
                ffmpeg_process.wait(timeout=5)
            except subprocess.TimeoutExpired:
                ffmpeg_process.kill()
                ffmpeg_process.wait()
            ffmpeg_process.communicate()
        raise

    finally:
        if writer:
            writer.close()
        if ffmpeg_process:
            if ffmpeg_process.poll() is None:
                ffmpeg_process.terminate()
                try:
                    ffmpeg_process.wait(timeout=5)
                except subprocess.TimeoutExpired:
                    ffmpeg_process.kill()
                    ffmpeg_process.wait()
            try:
                ffmpeg_process.communicate(timeout=5)
            except Exception:
                pass
        env.close()
        if use_virtual_display:
            if original_sdl is None:
                os.environ.pop("SDL_VIDEODRIVER", None)
            else:
                os.environ["SDL_VIDEODRIVER"] = original_sdl



from typing import Literal, Optional, Callable
from io import StringIO
from types import SimpleNamespace
import json

import os
import gymnasium as gym
import random
import string
import sys
import shutil

from rich.align import Align
from rich.text import Text

from sai_rl.api.routes.competition import CompetitionType
from sai_rl.api.routes.environment import EnvironmentType

from sai_rl.api.routes.package import PackageType
from sai_rl.package_control import PackageControl
from sai_rl.sai_console import SAIConsole, SAIStatus
from sai_rl.utils import TeeIO, config
from sai_rl.error import (
    BenchmarkError,
    CompetitionError,
    SubmissionError,
    SetupError,
    EnvironmentError,
)

from sai_rl.api import APIClient
from sai_rl.model import ModelManager
from sai_rl.types import ModelType, ModelLibraryType
from sai_rl.benchmark import (
    run_benchmark,
    record_episode,
    BenchmarkResults,
)


class SAIClient(object):
    """
    Main client for interacting with the SAI platform.

    The SAIClient provides methods for:
    - Managing competitions and submissions
    - Loading and evaluating models
    - Working with action functions
    - Managing environment packages
    - Running benchmarks and watching agents

    Args:
        env_id (Optional[str]): ID of environment to load
        api_key (Optional[str]): API key for authentication
        competition_id (Optional[str]): ID of competition to load
        action_function_id (Optional[str]): ID of action function to use
        api_base (Optional[str]): Custom API endpoint
        max_network_retries (Optional[int]): Max API retry attempts
        console (Optional[SAIConsole]): Custom console for output

    Examples:
        Basic usage:
        >>> client = SAIClient(api_key="your-api-key")
        >>> client.load_competition("comp-123456")
        >>> client.watch()  # Watch random agent

        Using a model:
        >>> model = torch.load("my_model.pt")
        >>> client.benchmark(model=model, model_type="pytorch")

        Working with action functions:
        >>> client.load_action_fn("./my_action.py")
        >>> client.submit_action_fn()
    """

    def __init__(
        self,
        env_id: Optional[str] = None,
        api_key: Optional[str] = None,
        *,
        comp_id: Optional[str] = None,
        api_base: Optional[str] = None,
        is_server: bool = False,
    ):
        if is_server:
            self._log_capture = StringIO()
            self._original_stdout = sys.stdout
            sys.stdout = TeeIO(self._log_capture, sys.__stdout__)
        else:
            self._log_capture = None
            self._original_stdout = None
        self._console = SAIConsole()

        with self._console.status("Loading SAI CLI...") as status:
            # Setup API client
            self._api = APIClient(
                console=self._console, api_key=api_key, api_base=api_base
            )

            # Setup package control
            self._package_control = PackageControl(
                api=self._api,
                console=self._console,
                is_disabled=is_server,
            )

            self._console.display_title(
                self._package_control._get_package_version("sai_rl") or "unknown",
                self._package_control._is_editable_install("sai_rl"),
            )

            self._package_control.setup(status)

            self._package: Optional[PackageType] = None
            self._environment: Optional[EnvironmentType] = None
            self._competition: Optional[CompetitionType] = None

            self._env = None

            if comp_id is not None:
                self._console.print()
                self._load_competition(comp_id, status)

            elif env_id is not None:
                self._console.print()
                self._load_environment(env_id, status)

            self._console.print()

    # ---- Internal Utility Methods ----
    def _check_setup(self) -> bool:
        if not self._package_control.setup_complete:
            raise SetupError("Setup not complete")
        assert self._package_control.setup_complete
        return True

    def _get_logs(self) -> str:
        if self._log_capture:
            return self._log_capture.getvalue()
        return ""

    # ---- Print Methods ----
    def _print_environment(self):
        if not self._environment:
            raise EnvironmentError("Environment not loaded")

        id = self._environment.get("gymId")
        name = self._environment.get("name")
        package = self._environment.get("package").get("name")
        gym_id = self._environment.get("gymId")
        gym_type = self._environment.get("type")
        gym_vars = {}

        link = f"{config.platform_url}/environments/{id}"
        if self._competition:
            id = self._competition.get("slug")
            link = f"{config.platform_url}/competitions/{self._competition.get('id')}"
            gym_vars = self._competition.get("environmentVariables")

        title = f'"{name}" ({id})'

        info_group = f"""[bold cyan]
Env:[/bold cyan]    {gym_id}
[bold cyan]Env Standard:[/bold cyan] {gym_type}
[bold cyan]Env Vars:[/bold cyan] {json.dumps(gym_vars, indent=2)}
[bold cyan]Package:[/bold cyan] {package}"""

        link_group = f"[link={link}]View in Platform →[/link]"

        env_info = self._console.group(
            Align.left(Text.from_markup(info_group)),
            Align.right(Text.from_markup(link_group)),
        )

        panel = self._console.panel(env_info, title=title, padding=(0, 2))
        self._console.print(panel)

    def _print_submission_details(
        self,
        name: str,
        model_manager: ModelManager,
    ):
        if not self._competition:
            raise CompetitionError("Competition not loaded")

        title = f'"{name}" Submission Details'

        info_group = f"""[bold cyan]
Competition ID:[/bold cyan]      {self._competition.get("id")}
[bold cyan]Competition Name:[/bold cyan]    {self._competition.get("name")}
[bold cyan]Model Type:[/bold cyan]          {model_manager.model_type}
[bold cyan]Action Function:[/bold cyan]  {"Custom" if model_manager._action_manager else f"Default ({'sample' if model_manager._handler.is_continuous else 'argmax'})"}"""

        submission_info = self._console.group(Align.left(Text.from_markup(info_group)))

        panel = self._console.panel(submission_info, title=title, padding=(0, 2))
        self._console.print(panel)

    # ---- Load Methods ----
    def _load_competition_from_json(
        self, competition_json: str, status: Optional[SAIStatus] = None
    ):
        if status:
            status.update("Loading competition...")

        self._package = None
        self._environment = None
        self._competition = None

        try:
            comp_env = json.loads(
                competition_json, object_hook=lambda d: SimpleNamespace(**d)
            )
            competition: CompetitionType = {
                "id": comp_env.id,
                "slug": comp_env.slug,
                "name": comp_env.name,
                "opensource": comp_env.opensource
                if hasattr(comp_env, "opensource")
                else False,
                "environment": {
                    "id": comp_env.environment.id,
                    "name": comp_env.environment.name,
                    "gymId": comp_env.environment.gymId,
                    "type": comp_env.environment.type,
                    "actionType": comp_env.environment.actionType,
                    "package": {
                        "id": comp_env.environment.package.id,
                        "name": comp_env.environment.package.name,
                        "version": comp_env.environment.package.version,
                    },
                },
                "environmentVariables": vars(comp_env.environmentVariables)
                if hasattr(comp_env, "environmentVariables")
                and not isinstance(comp_env.environmentVariables, dict)
                else (
                    comp_env.environmentVariables
                    if hasattr(comp_env, "environmentVariables")
                    else {}
                ),
                "numberOfBenchmarks": comp_env.numberOfBenchmarks
                if hasattr(comp_env, "numberOfBenchmarks")
                else None,
                "seed": comp_env.seed if hasattr(comp_env, "seed") else None,
            }

            self._competition = competition
            self._environment = self._competition.get("environment")
            self._package = self._environment.get("package")

            self._print_environment()

            package_name = self._package.get("name")
            if package_name:
                self._package_control.load(package_name, status=status)

            self._console.success("Successfully loaded competition.")
            return self._competition
        except json.JSONDecodeError as e:
            raise CompetitionError(f"Invalid JSON format: {e}")
        except Exception as e:
            raise CompetitionError(f"Failed to load competition: {e}")

    def _load_competition(
        self, competition_id: str, status: Optional[SAIStatus] = None
    ):
        if status:
            status.update(f"Loading competition {competition_id}...")

        self._package = None
        self._environment = None
        self._competition = None

        self._competition = self._api.competition.get(competition_id)
        if not self._competition:
            raise CompetitionError("Competition not found")

        self._environment = self._competition.get("environment")
        self._package = self._environment.get("package")

        self._print_environment()

        package_name = self._package.get("name")
        if package_name:
            self._package_control.load(package_name, status=status)

        self._console.success("Successfully loaded competition.")
        return self._competition

    def _load_environment(
        self, environment_id: str, status: Optional[SAIStatus] = None
    ):
        if status:
            status.update(f"Loading environment {environment_id}...")

        self._package = None
        self._environment = None
        self._competition = None

        self._environment = self._api.environment.get(environment_id)
        if not self._environment:
            raise EnvironmentError("Environment not found")

        self._print_environment()

        self._package = self._environment.get("package")

        package_name = self._package.get("name")
        if package_name:
            self._package_control.load(package_name, status=status)

        self._console.success("Successfully loaded environment.")
        return self._environment

    # ---- Make Methods ----
    def _create_env_factory(self):
        if self._environment is None:
            raise EnvironmentError("Environment not loaded")

        env_id = self._environment.get("gymId")
        env_type = self._environment.get("type")
        competition_vars = (
            self._competition.get("environmentVariables") if self._competition else {}
        )

        def env_factory(**kwargs):
            env_vars = {
                **(competition_vars or {}),
                "render_mode": "rgb_array",
                **kwargs,
            }

            if env_type == "gymnasium":
                return gym.make(env_id, **env_vars)
            elif env_type == "gym-v26":
                return gym.make("GymV26Environment-v0", env_id=env_id, **env_vars)
            elif env_type == "gym-v21":
                return gym.make("GymV21Environment-v0", env_id=env_id, **env_vars)
            elif env_type == "pufferlib":
                try:
                    from pufferlib.emulation import GymnasiumPufferEnv
                    from pufferlib.ocean import env_creator

                    return GymnasiumPufferEnv(
                        env_creator=env_creator(env_id),
                        env_kwargs=env_vars,
                    )
                except ImportError:
                    raise ImportError(
                        "Pufferlib is not installed. "
                        "Please install it using \"pip install 'sai_rl[pufferlib]'\"."
                    )
            else:
                raise EnvironmentError(f"Unsupported environment type: {env_type}")

        return env_factory

    def _make_new_env(
        self,
        *,
        render_mode: Literal["human", "rgb_array"] = "rgb_array",
        **kwargs,
    ):
        if self._environment is None:
            raise EnvironmentError("Environment not loaded")

        if self._env is not None:
            self._env.close()
            self._env = None

        env_id = self._environment.get("gymId")
        competition_vars = (
            self._competition.get("environmentVariables") if self._competition else {}
        )
        env_vars = {**(competition_vars or {}), "render_mode": render_mode, **kwargs}

        if self._competition and kwargs:
            self._console.warning(
                "Additional environment arguments will override competitions settings."
                "Your local environment may not match the competition configuration."
            )

        env_type = self._environment.get("type")

        if env_type == "gymnasium":
            env = gym.make(env_id, **env_vars)
        elif env_type == "gym-v26":
            env = gym.make("GymV26Environment-v0", env_id=env_id, **env_vars)
        elif env_type == "gym-v21":
            env = gym.make("GymV21Environment-v0", env_id=env_id, **env_vars)
        elif env_type == "pettingzoo":
            raise NotImplementedError(
                "PettingZoo environments are not supported yet. "
                "Please use the competition environment."
            )
        elif env_type == "pufferlib":
            try:
                from pufferlib.emulation import GymnasiumPufferEnv
                from pufferlib.ocean import env_creator

                return GymnasiumPufferEnv(
                    env_creator=env_creator(env_id),
                    env_kwargs=env_vars,
                )
            except ImportError:
                raise ImportError(
                    "Pufferlib is not installed. "
                    "Please install it using \"pip install 'sai_rl[pufferlib]'\"."
                )
        else:
            raise EnvironmentError(
                f"Unsupported environment type: {env_type}. "
                "Please use a supported environment."
            )

        self._env = env
        return env

    ############################################################
    # Public Methods
    ############################################################

    # ---- Competition Methods ----
    def get_competition(self):
        """
        Gets information about the currently loaded competition.

        Returns:
            dict: Competition information including:
                - id: Competition ID
                - name: Competition name
                - env_name: Environment name
                - env_lib: Environment library
                - env_vars: Environment variables
            None: If no competition is loaded

        Examples:
            >>> comp = client.get_competition()
            >>> if comp:
            ...     print(f"Using {comp['env_name']}")
        """
        self._check_setup()
        return self._competition

    def load_competition(self, competition_id: str):
        """
        Loads a competition by its ID.

        Args:
            competition_id (str): Platform ID of the competition to load

        Returns:
            dict: Loaded competition information

        Raises:
            CompetitionError: If competition cannot be loaded

        Examples:
            >>> client.load_competition("comp-123456")
            >>> client.watch()  # Watch random agent
        """
        self._check_setup()

        with self._console.status(f"Loading competition {competition_id}...") as status:
            return self._load_competition(competition_id, status=status)

    def list_competitions(self, show_table: bool = True):
        """
        Lists all available competitions.

        Args:
            show_table (bool): Whether to display a formatted table of competitions

        Returns:
            list[dict]: List of competition information including:
                - id: Competition ID
                - name: Competition name
                - description: Competition description
                - environment_name: Name of the environment
                - link: URL to view in platform

        Examples:
            >>> # Show table and get data
            >>> competitions = client.list_competitions()
            >>>
            >>> # Get data only
            >>> competitions = client.list_competitions(show_table=False)
        """
        self._check_setup()
        self._console.print()

        with self._console.status("Loading competitions...") as status:
            competitions = self._api.competition.list()
            if competitions is None:
                self._console.warning("No competitions found.")
                return []

            if show_table:
                status.update("Displaying competitions...")
                table = self._console.table("Available Competitions")

                table.add_column("ID", style="yellow")
                table.add_column("Name", style="cyan", no_wrap=True)
                table.add_column("Description", style="magenta")
                table.add_column("Environment", style="green")
                table.add_column("View in Platform", style="green", justify="center")

                for comp in competitions:
                    table.add_row(
                        comp.get("slug"),
                        comp.get("name"),
                        comp.get("description"),
                        comp.get("environment").get("name"),
                        f"[link={comp.get('link')}]View →[/link]",
                    )

                self._console.print(table)
                self._console.print()

            return competitions

    # ---- Submission Methods ----
    def list_submissions(self, show_table: bool = True):
        """
        Lists all your submissions across competitions.

        Args:
            show_table (bool): Whether to display a formatted table of submissions
                If True, shows a rich table with clickable platform links
                If False, returns data only

        Returns:
            list[dict]: List of submission information including:
                - id: Submission ID
                - name: Submission name
                - status: Current status ("pending", "running", "completed", "failed")
                - last_score: Most recent evaluation score
                - competition: Name of the competition
                - link: URL to view submission in platform

        Note: Returns empty list if no submissions are found

        Examples:
            >>> # Show table and get data
            >>> submissions = client.list_submissions()
            >>>
            >>> # Get data only
            >>> submissions = client.list_submissions(show_table=False)
            >>> for sub in submissions:
            ...     print(f"{sub['name']}: {sub['last_score']}")
            >>>
            >>> # Filter completed submissions
            >>> completed = [s for s in submissions
            ...             if s['status'] == 'completed']
            >>> print(f"Found {len(completed)} completed submissions")
        """
        self._check_setup()
        self._console.print()

        with self._console.status("Loading submissions...") as status:
            submissions = self._api.submission.list()
            if submissions is None:
                self._console.warning("No submissions found.")
                return []

            if show_table:
                status.update("Displaying submissions...")
                table = self._console.table("Your Submissions")
                table.add_column("ID", style="yellow")
                table.add_column("Name", style="cyan", no_wrap=True)
                table.add_column("Model Type", style="magenta")
                table.add_column("Competition", style="cyan")
                table.add_column("Environment", style="green")
                table.add_column("Last Score", style="bright_green", justify="right")
                table.add_column("SAIStatus", style="bright_yellow")
                table.add_column(
                    "View in Platform", style="bright_blue", justify="center"
                )

                for submission in submissions:
                    table.add_row(
                        submission.get("id"),
                        submission.get("name"),
                        submission.get("model_type"),
                        submission.get("competition_name"),
                        submission.get("environment_name"),
                        submission.get("last_score"),
                        submission.get("status"),
                        f"[link={submission.get('link')}]View →[/link]"
                        if submission.get("link")
                        else "",
                    )

                self._console.print(table)
                self._console.print()

            return submissions

    # ---- Environment Methods ----
    def get_environment(self):
        """
        Gets information about the currently loaded environment.

        Returns:
            dict: Environment information including:
                - id: Environment ID
                - slug: Environment slug
                - name: Environment name
                - env_name: Environment name in the registry
                - env_library: Environment library
            None: If no environment is loaded

        Examples:
            >>> env = client.get_environment()
            >>> if env:
            ...     print(f"Using {env['env_name']}")
        """
        self._check_setup()
        return self._environment

    def load_environment(self, environment_id: str):
        """
        Loads an environment by its ID.

        Args:
            environment_id (str): Platform ID of the environment to load

        Returns:
            dict: Loaded environment information

        Raises:
            EnvironmentError: If environment cannot be loaded

        Examples:
            >>> client.load_environment("env-123456")
            >>> client.watch()  # Watch random agent
        """
        self._check_setup()

        with self._console.status(f"Loading environment {environment_id}...") as status:
            return self._load_environment(environment_id, status=status)

    def list_environments(self, show_table: bool = True):
        """
        Lists all available environments.

        Args:
            show_table (bool): Whether to display a formatted table of environments

        Returns:
            list[dict]: List of environment information including:
                - id: Environment ID
                - name: Environment name
                - description: Environment description
                - link: URL to view in platform

        Examples:
            >>> # Show table and get data
            >>> environments = client.list_environments()
            >>>
            >>> # Get data only
            >>> environments = client.list_environments(show_table=False)
        """
        self._check_setup()
        self._console.print()

        with self._console.status("Loading environments...") as status:
            environments = self._api.environment.list()
            if environments is None:
                self._console.warning("No environments found.")
                return []

            if show_table:
                status.update("Displaying environments...")
                table = self._console.table("Available Environments")

                table.add_column("ID", style="yellow")
                table.add_column("Name", style="cyan", no_wrap=True)
                table.add_column("Description", style="magenta")
                table.add_column("View in Platform", style="green", justify="center")

                for env in environments:
                    table.add_row(
                        env.get("env_name"),
                        env.get("name"),
                        env.get("description"),
                        f"[link={env.get('link')}]View →[/link]",
                    )

                self._console.print(table)
                self._console.print()

            return environments

    def reset(self):
        """
        Clears the currently loaded environment.

        This removes all references to the current environment, allowing
        you to load a different one or start fresh.

        Examples:
            >>> client.reset_environment()
            >>> client.load_environment("new-env-id")
        """
        self._check_setup()
        self._competition = None
        self._environment = None
        self._package = None

    def make_env(self, render_mode: Literal["human", "rgb_array"] = "human", **kwargs):
        """
        Creates a new instance of the competition environment.

        Args:
            render_mode (Literal["human", "rgb_array"]): How to render the environment
                - "human": Display environment in a window
                - "rgb_array": Return RGB array for video recording
            **kwargs: Additional keyword arguments to pass to the environment
                Note: These will be ignored when using a competition environment

        Returns:
            gym.Env: A Gymnasium environment instance

        Raises:
            CompetitionError: If no competition is loaded

        Examples:
            >>> env = client.make_env()
            >>> obs, _ = env.reset()
            >>> env.render()

            >>> # With custom environment args (only works for non-competition environments)
            >>> env = client.make_env(truncate_episode_steps=100)
        """
        self._check_setup()
        return self._make_new_env(render_mode=render_mode, **kwargs)

    def watch(
        self,
        model: Optional[ModelType] = None,
        action_function: Optional[str | Callable] = None,
        *,
        model_type: Optional[ModelLibraryType] = None,
        algorithm: Optional[str] = None,
        num_runs: int = 1,
    ):
        """
        Watch a model (or random agent) interact with the environment.

        Args:
            model (Optional[ModelType]): Model to watch. Can be one of:
                - PyTorch model (torch.nn.Module)
                - TensorFlow model (tf.Module or tensorflow.compat.v1)
                - Keras model (tf.keras.Model)
                - Stable-Baselines3 model (BaseAlgorithm)
                - ONNX model (OnnxAgentWrapper)
                - URL or file path (str)
                If None, uses random actions
            model_type (Optional[ModelLibraryType]): Framework used. One of:
                - "pytorch"
                - "tensorflow"
                - "keras"
                - "stable_baselines3"
                - "onnx"
            runs (int): Number of episodes to run (default: 1)

        Raises:
            BenchmarkError: If watching fails
            CompetitionError: If no competition is loaded

        Examples:
            >>> # Watch random agent
            >>> client.watch()
            >>>
            >>> # Watch PyTorch model
            >>> model = torch.load("model.pt")
            >>> client.watch(model=model, model_type="pytorch", runs=3)
        """
        self._check_setup()
        if not self._environment:
            raise EnvironmentError("Environment not loaded")

        env = None
        self._console.print()
        with self._console.status("Setting up environment...") as status:
            try:
                env = self._make_new_env(render_mode="human")

                if model is not None:
                    model_manager = ModelManager(
                        env=env,
                        model=model,
                        model_type=model_type,
                        algorithm=algorithm,
                        action_function=action_function,
                        console=self._console,
                        status=status,
                    )
                    status.update(
                        f"Watching {model_manager.model_type} model in '{env.spec.id}' environment..."  # type: ignore
                    )
                else:
                    model_manager = None
                    status.update(
                        f"Watching random agent in '{env.spec.id}' environment..."  # type: ignore
                    )

                self._console.print()

                for run_index in range(num_runs):
                    self._console.info(
                        f"Running watch {run_index + 1} of {num_runs}..."
                    )

                    obs, _ = env.reset()
                    terminated = False
                    truncated = False
                    scores = 0
                    timesteps = 0

                    while not terminated and not truncated:
                        if model_manager:
                            action = model_manager.get_action(obs)[0]
                        else:
                            action = env.action_space.sample()

                        obs, reward, terminated, truncated, info = env.step(action)
                        env.render()

                        scores += reward  # type: ignore
                        timesteps += 1

                    env.close()

                    self._console.success(
                        f"Episode finished after {timesteps} timesteps with score: {scores}"
                    )

                self._console.print()

            except Exception as e:
                self._console.error(f"Unable to watch model: {e}")
                raise BenchmarkError(e)

            finally:
                if env:
                    env.close()

    def benchmark(
        self,
        model: Optional[ModelType] = None,
        action_function: Optional[str | Callable] = None,
        *,
        model_type: Optional[ModelLibraryType] = None,
        algorithm: Optional[str] = None,
        num_envs: Optional[int] = None,
        record: bool = False,
        video_dir: Optional[str] = "results",
        show_progress: bool = True,
        throw_errors: bool = True,
        timeout: int = 600,
    ) -> BenchmarkResults:
        """
        Run benchmark evaluation of a model.

        Args:
            model (Optional[ModelType]): Model to evaluate. Can be one of:
                - PyTorch model (torch.nn.Module)
                - TensorFlow model (tf.Module or tensorflow.compat.v1)
                - Keras model (tf.keras.Model)
                - Stable-Baselines3 model (BaseAlgorithm)
                - ONNX model (onnxruntime.InferenceSession)
                - URL or file path (str)
                If None, uses random actions
            model_type (Optional[ModelLibraryType]): Framework used. One of:
                - "pytorch"
                - "tensorflow"
                - "keras"
                - "stable_baselines3"
                - "onnx"
            video_dir (Optional[str]): Path to save video recording
                If None, no video is recorded
            show_progress (bool): Whether to show progress bar during benchmark
                Defaults to True
            throw_errors (bool): Whether to raise exceptions on errors
                If False, returns error in results instead
                Defaults to True

        Returns:
            BenchmarkResults: Results containing:
                - status: "success", "error", or "timeout"
                - score: Total reward achieved
                - duration: Time taken in seconds
                - error: Error message if failed

        Raises:
            BenchmarkError: If benchmark fails and throw_errors is True
            CompetitionError: If no competition is loaded

        Examples:
            >>> # Benchmark random agent
            >>> results = client.benchmark()
            >>> print(f"Score: {results['score']}")
            >>>
            >>> # Benchmark PyTorch model with video
            >>> model = torch.load("model.pt")
            >>> results = client.benchmark(
            ...     model=model,
            ...     model_type="pytorch",
            ...     video_path="benchmark.mp4"
            ... )
        """
        self._console.print()

        results: BenchmarkResults = {
            "status": "error",
            "score": None,
            "duration": 0,
            "matches": [],
            "logs": None,
            "error": None,
        }

        env = None
        with self._console.status("Setting up benchmark...") as status:
            try:
                self._check_setup()
                if not self._environment:
                    raise EnvironmentError("Environment not loaded")
            except Exception as e:
                results["status"] = "error"
                results["error"] = str(e)

                if throw_errors and results.get("status") == "error":
                    raise BenchmarkError(results.get("error"))
                return results

            try:
                status.update("Setting up environment...")
                env = self._make_new_env()

                if env is None:
                    raise EnvironmentError("Environment not loaded")

                render_mode = None
                if record:
                    # Check if the environment supports video recording
                    render_modes = env.metadata.get("render_modes", [])

                    if "rgb_array" in render_modes:
                        render_mode = "rgb_array"
                    elif "human" in render_modes:
                        render_mode = "human"
                    elif self._environment.get("type") == "pufferlib":
                        render_mode = "rgb_array"
                    else:
                        self._console.warning(
                            "Video recording not supported for this environment. "
                            "Continuing without video."
                        )

                model_manager = None
                if model is not None:
                    model_manager = ModelManager(
                        console=self._console,
                        env=env,
                        model=model,
                        model_type=model_type,
                        algorithm=algorithm,
                        action_function=action_function,
                        status=status,
                    )
                else:
                    self._console.warning("No model provided. Using random actions.")

                self._console.print()
                env.close()
                env = None
                self._env = None

                env_factory = self._create_env_factory()
                num_environments = (
                    num_envs
                    if num_envs is not None
                    else self._competition.get("numberOfBenchmarks")
                    if self._competition
                    and self._competition.get("numberOfBenchmarks") is not None
                    else 20
                )

                results, actions, seeds = run_benchmark(
                    env_creator=env_factory,
                    console=self._console,
                    seed=self._competition.get("seed") if self._competition else None,
                    model=model_manager,
                    num_envs=num_environments or 20,
                    status=status,
                    show_progress=show_progress,
                    save_actions=record,
                    timeout=timeout,
                )

                if seeds is not None:
                    scores = {
                        seed: results["matches"][i]["score"]
                        for i, seed in enumerate(seeds)
                    }

                    # Check if all scores are the same
                    unique_scores = set(scores.values())

                    seeds_to_record = {}
                    if len(unique_scores) == 1:
                        # All scores are identical - just pick different seeds for variety
                        seed_list = list(scores.keys())
                        seeds_to_record = {
                            "best": seed_list[0],
                            "worst": seed_list[min(1, len(seed_list) - 1)],
                            "average": seed_list[min(2, len(seed_list) - 1)],
                        }
                    else:
                        best_seed, _ = max(scores.items(), key=lambda x: x[1])
                        worst_seed, _ = min(scores.items(), key=lambda x: x[1])

                        # Find average score seed that's different from best and worst
                        used_seeds = {best_seed, worst_seed}
                        remaining_scores = [
                            (seed, score)
                            for seed, score in scores.items()
                            if seed not in used_seeds
                        ]

                        avg_score = sum(scores.values()) / len(scores)
                        average_seed, _ = min(
                            remaining_scores, key=lambda x: abs(x[1] - avg_score)
                        )

                        seeds_to_record = {
                            "best": best_seed,
                            "worst": worst_seed,
                            "average": average_seed,
                        }

                    should_record = record and render_mode and video_dir
                    if should_record and results["status"] == "timeout":
                        self._console.warning(
                            "The benchmark was interrupted by the timeout, match recordings will not be saved."
                        )

                    if should_record and results["status"] == "success":
                        assert render_mode is not None
                        assert video_dir is not None

                        if not os.path.exists(video_dir):
                            os.makedirs(video_dir)

                        for label, seed in seeds_to_record.items():
                            match_index = seeds.index(seed)
                            video_path = os.path.join(video_dir, f"{label}.mp4")
                            record_episode(
                                env_creator=env_factory,  # type: ignore
                                render_mode=render_mode,
                                seed=seed,
                                actions=actions[match_index],
                                output_path=video_path,
                            )
                            self._console.info(
                                f"✓ {label.capitalize()} episode finished recording."
                            )
                            results["matches"][match_index]["videoKey"] = label

                self._console.print()
                self._console.debug(
                    f"\n[bold]Results:[/bold]\n{json.dumps(results, indent=2)}"
                )

                if results["status"] == "success":
                    self._console.success("Benchmark completed successfully")

            except Exception as e:
                results["status"] = "error"
                results["error"] = str(e)

            finally:
                if env:
                    env.close()

                if self._log_capture:
                    results["logs"] = self._get_logs()

            if throw_errors and results.get("status") == "error":
                raise BenchmarkError(results.get("error"))

            self._console.print()
            return results

    # ---- Model Methods ----
    def submit_model(
        self,
        name: str,
        model: ModelType,
        action_function: Optional[str | Callable] = None,
        *,
        model_type: Optional[ModelLibraryType] = None,
        algorithm: Optional[str] = None,
        use_onnx: bool = False,
        tag: str = "default",
    ):
        return self.submit(
            name=name,
            model=model,
            action_function=action_function,
            model_type=model_type,
            algorithm=algorithm,
            use_onnx=use_onnx,
            tag=tag,
        )

    def submit(
        self,
        name: str,
        model: ModelType,
        action_function: Optional[str | Callable] = None,
        *,
        model_type: Optional[ModelLibraryType] = None,
        algorithm: Optional[str] = None,
        use_onnx: bool = False,
        tag: str = "default",
    ):
        """
        Submits a model to the current competition.

        Args:
            name (str): Name for the submission
            model (ModelType): Model to submit. Can be one of:
                - PyTorch model (torch.nn.Module)
                - TensorFlow model (tf.Module or tensorflow.compat.v1)
                - Keras model (tf.keras.Model)
                - Stable-Baselines3 model (BaseAlgorithm)
                - ONNX model (OnnxAgentWrapper)
                - URL or file path (str)
            model_type (Optional[ModelLibraryType]): Framework used. One of:
                - "pytorch" - For PyTorch models
                - "tensorflow" - For TensorFlow v1 and v2 models
                - "keras" - For Keras models
                - "stable_baselines3" - For Stable-Baselines3 models
                - "onnx" - For ONNX models
            action_function_id (Optional[str]): ID of a previously submitted action function to use with this model
            use_onnx (bool): Whether to convert the model to ONNX format before submission
            skip_warning (bool): Skip validation warnings related to action function configuration

        Returns:
            dict: Submission information including:
                - id: Unique submission ID
                - name: Submission name
                - type: Model framework type
                - status: Current submission status
                - created_at: Timestamp of submission
                - updated_at: Timestamp of last update

        Raises:
            SubmissionError: If submission fails validation checks or upload fails
            CompetitionError: If no competition is currently loaded
            ValueError: If model type is invalid or incompatible with provided model

        Examples:
            >>> # Submit PyTorch model
            >>> model = torch.load("model.pt")
            >>> result = client.submit_model(
            ...     name="My Model v1",
            ...     model=model,
            ...     model_type="pytorch"
            ... )
            >>>
            >>> # Submit with action function and ONNX conversion
            >>> client.load_action_fn("./action.py")
            >>> action_fn = client.submit_action_fn()
            >>> result = client.submit_model(
            ...     name="My Model v2",
            ...     model=model,
            ...     model_type="pytorch",
            ...     action_function_id=action_fn["id"],
            ...     use_onnx=True
            ... )
        """
        self._check_setup()
        if not self._competition:
            raise CompetitionError(
                "No competition is loaded, please load a competition first using SAIClient(competition_id='')"
            )

        if self._competition["opensource"]:
            raise SubmissionError(
                "Cannot use python to submit to an open source competition. Please submit through the SAI website."
            )

        with self._console.status("Submitting model to the competition...") as status:
            env = self._make_new_env()

            model_manager = ModelManager(
                console=self._console,
                env=env,
                model=model,
                model_type=model_type,
                algorithm=algorithm,
                action_function=action_function,
            )

            self._console.print()

            self._print_submission_details(
                name,
                model_manager,  # use_onnx, action_function_id
            )

            file_extension = {
                "stable_baselines3": ".zip",
                "pytorch": ".pt",
                "tensorflow": ".pb",
                "keras": ".keras",
                "onnx": ".onnx",
            }.get("onnx" if use_onnx else model_manager.model_type, "")

            isTensorflowV2 = False
            if model_manager.model_type == "tensorflow" and not use_onnx:
                isTensorflowV2 = model_manager._handler.is_tf2_model(model)

            random_id = "".join(
                random.choices(string.ascii_lowercase + string.digits, k=6)
            )
            os.makedirs(config.temp_path, exist_ok=True)
            temp_model_path = f"{config.temp_path}/{random_id}{file_extension if not isTensorflowV2 else ''}"
            model_manager.save_model(temp_model_path, use_onnx=use_onnx)

            if action_function:
                temp_action_fn_path = f"{config.temp_path}/{random_id}.py"
                model_manager.save_action_function(temp_action_fn_path)

            status.update("Creating submission...")

            adj_temp_model_path = (
                f"{temp_model_path}{'/saved_model.pb' if isTensorflowV2 else ''}"
            )
            files = {
                "model": (
                    os.path.basename(adj_temp_model_path),
                    open(adj_temp_model_path, "rb"),
                    "application/octet-stream",
                ),
            }

            if action_function:
                files["actionFunction"] = (
                    os.path.basename(temp_action_fn_path),
                    open(temp_action_fn_path, "rb"),
                    "text/plain",
                )

            submission = None
            try:
                submission = self._api.submission.create(
                    {
                        "name": name,
                        "type": model_manager.model_type,
                        "competitionId": self._competition.get("id"),
                        "algorithm": model_manager.algorithm,
                        "method": "python",
                        "tag": tag,
                    },
                    files=files,
                )
            finally:
                for file_tuple in files.values():
                    file_tuple[1].close()

            if not submission:
                raise SubmissionError("Failed to create submission")

            if os.path.exists(temp_model_path):
                if isTensorflowV2 and os.path.isdir(temp_model_path):
                    shutil.rmtree(temp_model_path)  # remove directory
                else:
                    os.remove(temp_model_path)  # remove file
            else:
                self._console.warning("Temporary model file not found.")

            self._console.success("Model submitted successfully.")

            return {
                "id": submission.get("id"),
                "name": submission.get("name"),
                "type": submission.get("type"),
                "status": submission.get("status"),
            }

    def save_model(
        self,
        name: str,
        model: ModelType,
        model_type: Optional[ModelLibraryType] = None,
        algorithm: Optional[str] = None,
        use_onnx: bool = False,
        output_path: str = "./",
    ):
        """
        Saves a model to disk in the appropriate format.

        Args:
            name (str): Name for the saved model file (without extension)
            model (ModelType): Model to save. Can be one of:
                - PyTorch model (torch.nn.Module)
                - TensorFlow model (tf.Module or tensorflow.compat.v1)
                - Keras model (tf.keras.Model)
                - Stable-Baselines3 model (BaseAlgorithm)
                - ONNX model (OnnxAgentWrapper)
                - URL or file path (str)
            model_type (Optional[ModelLibraryType]): Framework used. One of:
                - "pytorch" (.pt)
                - "tensorflow" (.pb)
                - "keras" (.keras)
                - "stable_baselines3" (.zip)
                - "onnx" (.onnx)
            use_onnx (bool): Whether to convert and save model in ONNX format (default: False)
            output_path (str): Directory to save the model file (default: "./")

        Returns:
            str: Full path to the saved model file

        Raises:
            ModelError: If model cannot be saved or converted to ONNX
            CompetitionError: If no competition is loaded

        Note:
            - File extension is automatically added based on model_type
            - If use_onnx=True, model will be converted and saved in ONNX format regardless of original type

        Examples:
            >>> # Save PyTorch model
            >>> model = torch.load("model.pt")
            >>> path = client.save_model(
            ...     name="my_model",
            ...     model=model,
            ...     model_type="pytorch",
            ...     output_path="./models"
            ... )  # Saves to ./models/my_model.pt
            >>> print(path)
            './models/my_model.pt'

            >>> # Save model in ONNX format
            >>> path = client.save_model(
            ...     name="my_model",
            ...     model=model,
            ...     model_type="pytorch",
            ...     use_onnx=True,
            ...     output_path="./models"
            ... )  # Saves to ./models/my_model.onnx
        """
        self._check_setup()
        self._console.print()

        with self._console.status("Setting up model...") as status:
            env = self._make_new_env()

            model_manager = ModelManager(
                console=self._console,
                env=env,
                model=model,
                model_type=model_type,
                algorithm=algorithm,
            )

            file_extension = {
                "stable_baselines3": ".zip",
                "pytorch": ".pt",
                "tensorflow": ".pb",
                "keras": ".keras",
                "onnx": ".onnx",
            }.get("onnx" if use_onnx else model_manager.model_type, "")

            save_path = f"{output_path}/{name}{file_extension}"
            os.makedirs(output_path, exist_ok=True)

            status.update("Saving model...")
            model_manager.save_model(save_path, use_onnx=use_onnx)

            self._console.success(f"Model saved to {save_path}")

        return save_path

    # ---- Package Methods ----
    def get_package(self, package: str) -> str:
        """
        Gets information about a specific package.

        Args:
            package (str): Name of the package to get information about

        Returns:
            PackageType: Package information including:
                - id: Package ID
                - name: Package name
                - description: Package description
                - version: Package version
            None: If package is not found

        Examples:
            >>> info = client.get_package("sai-pygame")
            >>> if info:
            ...     print(f"Latest version: {info['version']}")
        """
        self._check_setup()

        return self._api.package.get(package)  # type: ignore

    def update_package(self, package: str) -> None:
        """
        Updates a package to its latest version.

        Args:
            package (str): Name of the package to update

        Raises:
            PackageError: If the package cannot be updated

        Note: If package is an editable install, update will be skipped

        Examples:
            >>> client.update_package("sai-pygame")
        """
        self._check_setup()

        with self._console.status("Updating package...") as status:
            self._package_control.update(package, status)

    def install_package(self, package: str) -> None:
        """
        Installs a package if not already installed.

        Args:
            package (str): Name of the package to install

        Raises:
            PackageError: If the package cannot be installed

        Examples:
            >>> client.install_package("sai-pygame")
        """
        self._check_setup()

        with self._console.status("Installing package...") as status:
            self._package_control.update(package, status)

    def uninstall_package(self, package: str) -> None:
        """
        Uninstalls a package.

        Args:
            package (str): Name of the package to uninstall

        Raises:
            PackageError: If the package cannot be uninstalled

        Examples:
            >>> client.uninstall_package("sai-pygame")
        """
        self._check_setup()

        with self._console.status("Uninstalling package...") as status:
            self._package_control.uninstall(package, status)

    def list_packages(self, show_table: bool = True):
        """
        Lists all available packages and their installation status.

        Args:
            show_table (bool): Whether to display a formatted table of packages

        Returns:
            list[PackageType]: List of package information including:
                - id: Package ID
                - name: Package name
                - description: Package description
                - version: Latest version available

                Note: When show_table=True, displays additional information:
                    - Installed Version: Currently installed version
                    - Latest Version: Latest available version
                    - SAIStatus: Up to date/Update available/Not installed
                    - Install Type: Editable/Regular/Not installed

        Examples:
            >>> # Show table and get data
            >>> packages = client.list_packages()
            >>>
            >>> # Get data only
            >>> packages = client.list_packages(show_table=False)
            >>> for pkg in packages:
            ...     print(f"{pkg['name']}: {pkg['version']}")
        """
        self._check_setup()
        self._console.print()

        with self._console.status("Loading packages...") as status:
            packages = self._api.package.list()
            if not packages:
                self._console.warning("No packages found.")
                return []

            if show_table:
                status.update("Displaying packages...")
                table = self._console.table("Available Packages")

                table.add_column("Name", style="cyan")
                table.add_column("Installed Version", style="yellow")
                table.add_column("Latest Version", style="green")
                table.add_column("Status", style="magenta")
                table.add_column("Install Type", style="blue")

                for package in packages:
                    latest_version = package.get("version")
                    installed_version = self._package_control._get_package_version(
                        package["name"]
                    )

                    is_latest = latest_version == installed_version
                    is_installed = installed_version is not None
                    is_editable = self._package_control._is_editable_install(
                        package["name"]
                    )

                    status = (
                        "Up to date"
                        if is_latest
                        else "Update available"
                        if is_installed
                        else "Not installed"
                    )

                    install_type = (
                        "Editable"
                        if is_editable
                        else "Regular"
                        if is_installed
                        else "—"
                    )

                    table.add_row(
                        package.get("name"),
                        installed_version or "Not installed",
                        latest_version or "Unknown",
                        status,
                        install_type,
                    )

                self._console.print(table)
                self._console.print()

            return packages
